# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin import list settings configuration.
"""


from __future__ import annotations

import re

from datetime import datetime
from logging import getLogger
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)

from buildarr.config import RemoteMapEntry
from buildarr.state import state
from buildarr.types import BaseEnum, InstanceName, NonEmptyStr, Password
from pydantic import AnyHttpUrl, ConstrainedStr, Field, PositiveInt, validator
from typing_extensions import Annotated, Self

from ..api import api_delete, api_get, api_post, api_put
from ..secrets import SonarrSecrets
from ..types import SonarrApiKey
from .types import SonarrConfigBase, TraktAuthUser
from .util import trakt_expires_encoder

if TYPE_CHECKING:
    from typing import Collection

    from ..config import SonarrInstanceConfig


logger = getLogger(__name__)


class YearRange(ConstrainedStr):
    """
    Constrained string type for a singular year or range of years.
    """

    regex = re.compile(r"[0-9]+(?:-[0-9]+)?")

    # TODO: validate that the end year is higher than the start year


class Monitor(BaseEnum):
    """
    Enumeration for what kind of monitoring should be done in an import list.
    """

    all_episodes = "all"
    future_episodes = "future"
    missing_episodes = "missing"
    existing_episodes = "existing"
    pilot_episode = "pilot"
    first_season = "firstSeason"
    only_latest_season = "latestSeason"
    none = "none"


class SeriesType(BaseEnum):
    """
    Series type to classify media from an import list.
    """

    standard = "standard"
    daily = "daily"
    anime = "anime"


class TraktPopularListType(BaseEnum):
    """
    Types of popularity list in Trakt.
    """

    trending = 0
    popular = 1
    anticipated = 2
    top_watched_by_week = 3
    top_watched_by_month = 4
    top_watched_by_year = 5
    top_watched_by_alltime = 6
    recommended_by_week = 7
    recommended_by_month = 8
    recommended_by_year = 9
    recommended_by_alltime = 10


class TraktUserListType(BaseEnum):
    """
    Types of user list in Trakt.
    """

    user_watch_list = 0
    user_watched_list = 1
    user_collection_list = 2


class ImportList(SonarrConfigBase):
    """
    For more information on how an import list should be setup,
    refer to this guide on [WikiArr](https://wiki.servarr.com/en/sonarr/settings#import-lists).

    All import list types can have the following attributes configured.
    """

    enable_automatic_add: bool = True
    """
    Automatically add series to Sonarr upon syncing.
    """

    monitor: Monitor = Monitor.all_episodes
    """
    Define how Sonarr should monitor existing and new episodes of series.

    Values:

    * `all-episodes` - Monitor all episodes except specials
    * `future-episodes` - Monitor episodes that have not aired yet
    * `missing-episodes` - Monitor episodes that do not have files or have not aired yet
    * `existing-episodes` - Monitor episodes that have files or have not aired yet
    * `pilot-episode` - Monitor only the pilot episode of a series
    * `first-season` - Monitor all episodes of the first season (all other seasons will be ignored)
    * `only-latest-season` - Monitor all episodes of the latest season and future seasons
    * `none` - No episodes will be monitored
    """

    root_folder: NonEmptyStr
    """
    The root folder to add list items to.

    This attribute is required.
    """

    quality_profile: NonEmptyStr
    """
    The name of the quality profile list items will be added with.

    This attribute is required.
    """

    language_profile: NonEmptyStr
    """
    The name of the language profile list items will be added with.

    This attribute is required.
    """

    series_type: SeriesType = SeriesType.standard
    """
    The type of series that get imported from this import list.
    This option affects how Sonarr handles the media, such as renaming.

    Values:

    * `standard` - Episodes released with an `SxxEyy` pattern
    * `daily` - Episodes released daily or less frequently that use year-month-day (2017-05-25)
    * `anime` - Episodes released using an absolute episode number
    """

    season_folder: bool = True
    """
    Use a season folder for series imported from this import list.
    """

    tags: Set[NonEmptyStr] = set()
    """
    Tags to assign to items imported from this import list.
    """

    _implementation_name: str
    _implementation: str
    _config_contract: str
    _remote_map: List[RemoteMapEntry]

    @classmethod
    def _get_base_remote_map(
        cls,
        quality_profile_ids: Mapping[str, int],
        language_profile_ids: Mapping[str, int],
        tag_ids: Mapping[str, int],
    ) -> List[RemoteMapEntry]:
        """
        Return the remote map for the base import list attributes.

        Args:
            quality_profile_ids (Mapping[str, int]): Quality profile ID mapping on the remote.
            language_profile_ids (Mapping[str, int]): Language profile ID mapping on the remote.
            tag_ids (Mapping[str, int]): Tag ID mapping on the remote.

        Returns:
            Remote map (as a list of entries)
        """
        return [
            ("enable_automatic_add", "enableAutomaticAdd", {}),
            ("monitor", "shouldMonitor", {}),
            (
                "root_folder",
                "rootFolderPath",
                {"decoder": lambda v: v or None, "encoder": lambda v: v or ""},
            ),
            (
                "quality_profile",
                "qualityProfileId",
                {
                    "decoder": lambda v: next(
                        (ind for ind, ind_id in quality_profile_ids.items() if ind_id == v),
                    ),
                    "encoder": lambda v: quality_profile_ids[v],
                },
            ),
            (
                "language_profile",
                "languageProfileId",
                {
                    "decoder": lambda v: next(
                        (ind for ind, ind_id in language_profile_ids.items() if ind_id == v),
                    ),
                    "encoder": lambda v: language_profile_ids[v],
                },
            ),
            ("series_type", "seriesType", {}),
            ("season_folder", "seasonFolder", {}),
            (
                "tags",
                "tags",
                {
                    "decoder": lambda v: set(
                        (tag for tag, tag_id in tag_ids.items() if tag_id in v),
                    ),
                    "encoder": lambda v: sorted(tag_ids[tag] for tag in v),
                },
            ),
        ]

    @classmethod
    def _from_remote(
        cls,
        quality_profile_ids: Mapping[str, int],
        language_profile_ids: Mapping[str, int],
        tag_ids: Mapping[str, int],
        remote_attrs: Mapping[str, Any],
    ) -> Self:
        """
        Parse an import list object from the remote Sonarr instance,
        and return its internal representation.

        Args:
            quality_profile_ids (Mapping[str, int]): Quality profile ID mapping on the remote.
            language_profile_ids (Mapping[str, int]): Language profile ID mapping on the remote.
            tag_ids (Mapping[str, int]): Tag ID mapping on the remote.
            remote_attrs (Mapping[str, Any]): Remote instance import list object.

        Returns:
            Internal import list object
        """
        return cls(
            **cls.get_local_attrs(
                (
                    cls._get_base_remote_map(quality_profile_ids, language_profile_ids, tag_ids)
                    + cls._remote_map
                ),
                remote_attrs,
            ),
        )

    def _resolve(self, name: str, ignore_nonexistent_ids: bool = False) -> Self:
        """
        Resolve any instance references on this import list, and
        return an object with fully qualified attribute values.

        Args:
            name (str): Name associated with this import list.
            ignore_nonexistent_ids (bool, optional): Ignore invalid IDs on the target instance.

        Returns:
            Fully qualified import list object
        """
        return self._resolve_from_local(
            name=name,
            local=self,
            ignore_nonexistent_ids=ignore_nonexistent_ids,
        )

    def _resolve_from_local(
        self,
        name: str,
        local: Self,
        ignore_nonexistent_ids: bool = False,
    ) -> Self:
        """
        Resolve this import list using instance references from the given object,
        and return an object with fully qualified attribute values.

        Used to fully qualify import list objects read from a remote Sonarr instance,
        using its corresponding local configuration.

        Args:
            name (str): Name associated with this import list.
            local (Self): Import list object to use for instance referencing.
            ignore_nonexistent_ids (bool, optional): Ignore invalid IDs on the target instance.

        Returns:
            Fully qualified import list object
        """
        return self

    def _create_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        quality_profile_ids: Mapping[str, int],
        language_profile_ids: Mapping[str, int],
        tag_ids: Mapping[str, int],
        importlist_name: str,
    ) -> None:
        """
        Create this import list on the remote Sonarr instance.

        Args:
            tree (str): Configuration tree. Used for logging.
            secrets (SonarrSecrets): Secrets metadata for the remote instance.
            quality_profile_ids (Mapping[str, int]): Quality profile ID mapping on the remote.
            language_profile_ids (Mapping[str, int]): Language profile ID mapping on the remote.
            tag_ids (Mapping[str, int]): Tag ID mapping on the remote.
            importlist_name (str): Name associated with this import list.
        """
        api_post(
            secrets,
            "/api/v3/importlist",
            {
                "name": importlist_name,
                "implementation": self._implementation,
                "implementationName": self._implementation_name,
                "configContract": self._config_contract,
                **self.get_create_remote_attrs(
                    tree,
                    (
                        self._get_base_remote_map(
                            quality_profile_ids,
                            language_profile_ids,
                            tag_ids,
                        )
                        + self._remote_map
                    ),
                ),
            },
        )

    def _update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        quality_profile_ids: Mapping[str, int],
        language_profile_ids: Mapping[str, int],
        tag_ids: Mapping[str, int],
        importlist_id: int,
        importlist_name: str,
    ) -> bool:
        """
        Compare this import list to the currently active one the remote Sonarr instance,
        and update it in-place if there are differences.

        Args:
            tree (str): Configuration tree. Used for logging.
            secrets (SonarrSecrets): Secrets metadata for the remote instance.
            remote (Self): Active import list confiuration on the remote instance.
            quality_profile_ids (Mapping[str, int]): Quality profile ID mapping on the remote.
            language_profile_ids (Mapping[str, int]): Language profile ID mapping on the remote.
            tag_ids (Mapping[str, int]): Tag ID mapping on the remote.
            importlist_id (int): ID associated with this import list on the remote instance.
            importlist_name (str): Name associated with this import list.

        Returns:
            `True` if the import list was updated, otherwise `False`
        """
        updated, remote_attrs = self.get_update_remote_attrs(
            tree,
            remote,
            self._get_base_remote_map(quality_profile_ids, language_profile_ids, tag_ids)
            + self._remote_map,
            check_unmanaged=True,
            set_unchanged=True,
        )
        if updated:
            api_put(
                secrets,
                f"/api/v3/importlist/{importlist_id}",
                {
                    "id": importlist_id,
                    "name": importlist_name,
                    "implementation": self._implementation,
                    "implementationName": self._implementation_name,
                    "configContract": self._config_contract,
                    **remote_attrs,
                },
            )
            return True
        return False

    def _delete_remote(self, tree: str, secrets: SonarrSecrets, importlist_id: int) -> None:
        """
        Delete this import list from the remote Sonarr instance.

        Args:
            tree (str): Configuration tree. Used for logging.
            secrets (SonarrSecrets): Secrets metadata for the remote instance.
            importlist_id (int): ID associated with this import list on the remote instance.
        """
        logger.info("%s: (...) -> (deleted)", tree)
        api_delete(secrets, f"/api/v3/importlist/{importlist_id}")


class ProgramImportList(ImportList):
    """
    Base class for program-based import lists.
    """

    pass


class PlexImportList(ImportList):
    """
    Base class for import lists based on Plex.
    """

    pass


class TraktImportList(ImportList):
    """
    Import added media from a list on the Trakt media tracker.

    !!! note

        Sonarr directly authenticates with Trakt to generate tokens for it to use.
        At the moment, the easiest way to generate the tokens for Buildarr
        is to do it using the GUI within Sonarr, and use the following
        shell command to retrieve the generated configuration.

        ```bash
        $ curl -X "GET" "<sonarr-url>/api/v3/notification" -H "X-Api-Key: <api-key>"
        ```

    The following parameters are common to all Trakt import list types.
    The authenticated-related parameters (`access_token`, `refresh_token`, `expires`, `auth_user`)
    are required.
    """

    # Base class for import lists based on Trakt.

    # FIXME: Determine easier procedure for getting access tokens and test.

    access_token: Password
    """
    Access token for Sonarr from Trakt.
    """

    refresh_token: Password
    """
    Refresh token for Sonarr from Trakt.
    """

    expires: datetime
    """
    Expiry date-time of the access token, preferably in ISO-8601 format and in UTC.

    Example: `2023-05-10T15:34:08.117451Z`
    """

    auth_user: TraktAuthUser
    """
    The username being authenticated in Trakt.
    """

    # rating
    # TODO: constraints
    rating: NonEmptyStr = "0-100"  # type: ignore[assignment]
    """
    Filter series by rating range, with a maximum range of 0-100.
    """

    username: Optional[str] = None
    """
    Username for the list to import from.

    Leave undefined, empty or set to `None` to use the auth user.
    """

    genres: Set[NonEmptyStr] = set()
    """
    Filter series by Trakt genre slug.
    """

    years: Optional[YearRange] = None
    """
    Filter series by year or year range. (e.g. `2009` or `2009-2015`)
    """

    limit: PositiveInt = 100
    """
    Limit the number of series to get.
    """

    trakt_additional_parameters: Optional[str] = None
    """
    Additional parameters to send to the Trakt API.
    """

    @classmethod
    def _get_base_remote_map(
        cls,
        quality_profile_ids: Mapping[str, int],
        language_profile_ids: Mapping[str, int],
        tag_ids: Mapping[str, int],
    ) -> List[RemoteMapEntry]:
        return [
            *super()._get_base_remote_map(quality_profile_ids, language_profile_ids, tag_ids),
            ("access_token", "accessToken", {"is_field": True}),
            ("refresh_token", "refreshToken", {"is_field": True}),
            ("expires", "expires", {"is_field": True, "encoder": trakt_expires_encoder}),
            ("auth_user", "authUser", {"is_field": True}),
            ("rating", "rating", {"is_field": True}),
            (
                "username",
                "username",
                {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
            ),
            (
                "genres",
                "genres",
                {
                    "is_field": True,
                    "decoder": lambda v: set(v.split(",")) if v else set(),
                    "encoder": lambda v: ",".join(sorted(v)) if v else "",
                },
            ),
            (
                "years",
                "years",
                {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
            ),
            ("limit", "limit", {"is_field": True}),
            ("trakt_additional_parameters", "traktAdditionalParameters", {"is_field": True}),
        ]


class SonarrImportList(ProgramImportList):
    """
    Import items from another Sonarr instance.

    The linked Sonarr instance must be the same major version as this defined Sonarr instance.
    For example, a Sonarr V3 instance cannot connect with a Sonarr V4 instance, and vice versa.

    ```yaml
    ...
      import_lists:
        definitions:
          Sonarr:
            type: "sonarr"
            # Global import list options.
            root_folder: "/path/to/videos"
            quality_profile: "HD/SD"
            language_profile: "English"
            # Sonarr import list-specific options.
            full_url: "http://sonarr:8989"
            api_key: "1a2b3c4d5e1a2b3c4d5e1a2b3c4d5e1a"
            source_quality_profiles:
              - 11
              ...
            source_language_profiles:
              - 22
              ...
            source_tags:
              - 33
              ...
    ```

    This import list supports instance references to another Buildarr-defined Sonarr instance
    using `instance_name`.

    In this mode, you can specify `instance_name` in place of `api_key`,
    and use actual names for the source language profiles, quality profiles and tags,
    instead of IDs which are subject to change.

    Here is an example of one Sonarr instance (`sonarr-4k`) referencing
    another instance (`sonarr-hd`), using it as an import list.

    ```yaml
    sonarr:
      instances:
        sonarr-hd:
          hostname: "localhost"
          port: 8989
        sonarr-4k:
          hostname: "localhost"
          port: 8990
          settings:
            import_lists:
              definitions:
                Sonarr (HD):
                  type: "sonarr"
                  # Global import list options.
                  root_folder: "/path/to/videos"
                  quality_profile: "4K"
                  language_profile: "English"
                  # Sonarr import list-specific options.
                  full_url: "http://sonarr:8989"
                  instance_name: "sonarr-hd"
                  source_quality_profiles:
                    - "HD/SD"
                  source_language_profiles:
                    - "English"
                  source_tags:
                    - "shows"
    ```

    An important thing to keep in mind is that unless Buildarr is on the same network
    as the rest of the *Arr stack, the hostnames and ports may differ to what the
    Sonarr instances will use to communicate with each other. `full_url` should be
    set to what the Sonarr instance itself will use to link to the target instance.
    """

    type: Literal["sonarr"] = "sonarr"
    """
    Type value associated with this kind of import list.
    """

    instance_name: Optional[InstanceName] = Field(None, plugin="sonarr")
    """
    The name of the Sonarr instance within Buildarr, if linking this Sonarr instance
    with another Buildarr-defined Sonarr instance.

    *New in version 0.3.0.*
    """

    full_url: AnyHttpUrl
    """
    The URL that this Sonarr instance will use to connect to the source Sonarr instance.
    """

    api_key: Optional[SonarrApiKey] = None
    """
    API key used to access the source Sonarr instance.

    If a Sonarr instance managed by Buildarr is not referenced using `instance_name`,
    this attribute is required.
    """

    source_quality_profiles: Set[Union[PositiveInt, NonEmptyStr]] = Field(
        set(),
        alias="source_quality_profile_ids",
    )
    """
    List of IDs (or names) of the quality profiles on the source instance to import from.

    Quality profile names can only be used if `instance_name` is used to
    link to a Buildarr-defined Sonarr instance.
    If linking to a Sonarr instance outside Buildarr, IDs must be used.

    *Changed in version 0.3.0*: Renamed from `source_quality_profile_ids`
    (which is still valid as an alias), and added support for quality profile names.
    """

    source_language_profiles: Set[Union[PositiveInt, NonEmptyStr]] = Field(
        set(),
        alias="source_language_profile_ids",
    )
    """
    List of IDs (or names) of the language profiles on the source instance to import from.

    Language profile names can only be used if `instance_name` is used to
    link to a Buildarr-defined Sonarr instance.
    If linking to a Sonarr instance outside Buildarr, IDs must be used.

    *Changed in version 0.3.0*: Renamed from `source_language_profile_ids`
    (which is still valid as an alias), and added support for language profile names.
    """

    source_tags: Set[Union[PositiveInt, NonEmptyStr]] = Field(set(), alias="source_tag_ids")
    """
    List of IDs (or names) of the tags on the source instance to import from.

    Tag names can only be used if `instance_name` is used to
    link to a Buildarr-defined Sonarr instance.
    If linking to a Sonarr instance outside Buildarr, IDs must be used.

    *Changed in version 0.3.0*: Renamed from `source_tag_ids`
    (which is still valid as an alias), and added support for tag names.
    """

    _implementation_name: str = "Sonarr"
    _implementation: str = "SonarrImport"
    _config_contract: str = "SonarrSettings"
    _remote_map: List[RemoteMapEntry] = []

    @classmethod
    def _get_base_remote_map(
        cls,
        quality_profile_ids: Mapping[str, int],
        language_profile_ids: Mapping[str, int],
        tag_ids: Mapping[str, int],
    ) -> List[RemoteMapEntry]:
        return [
            *super()._get_base_remote_map(quality_profile_ids, language_profile_ids, tag_ids),
            ("full_url", "baseUrl", {"is_field": True}),
            ("api_key", "apiKey", {"is_field": True}),
            (
                "source_quality_profiles",
                "profileIds",
                {
                    "is_field": True,
                    "root_encoder": lambda vs: cls._encode_source_resources(
                        instance_name=vs.instance_name,
                        resources=vs.source_quality_profiles,
                        resource_type="qualityprofile",
                    ),
                },
            ),
            (
                "source_language_profiles",
                "languageProfileIds",
                {
                    "is_field": True,
                    "root_encoder": lambda vs: cls._encode_source_resources(
                        instance_name=vs.instance_name,
                        resources=vs.source_language_profiles,
                        resource_type="languageprofile",
                    ),
                },
            ),
            (
                "source_tags",
                "tagIds",
                {
                    "is_field": True,
                    "root_encoder": lambda vs: cls._encode_source_resources(
                        instance_name=vs.instance_name,
                        resources=vs.source_tags,
                        resource_type="tag",
                        name_key="label",
                    ),
                },
            ),
        ]

    @classmethod
    def _get_secrets(cls, instance_name: str) -> SonarrSecrets:
        """
        Fetch the secrets metadata for the given Sonarr instance from the Buildarr state.

        Args:
            instance_name (str): Name of Sonarr instance to get the secrets for.

        Returns:
            Sonarr instance secrets metadata
        """
        return cast(SonarrSecrets, state.secrets.sonarr[instance_name])

    @classmethod
    def _get_resources(cls, instance_name: str, resource_type: str) -> List[Dict[str, Any]]:
        """
        Make an API request to Sonarr to get the list of resources of the requested type.

        Args:
            instance_name (str): Name of Sonarr instance to get the resources from.
            profile_type (str): Name of the resource to get in the Sonarr API.

        Returns:
            List of resource API objects
        """
        return api_get(cls._get_secrets(instance_name), f"/api/v3/{resource_type}")

    @validator("api_key", always=True)
    def validate_api_key(
        cls,
        value: Optional[SonarrApiKey],
        values: Mapping[str, Any],
    ) -> Optional[SonarrApiKey]:
        """
        Validate the `api_key` attribute after parsing.

        Args:
            value (Optional[str]): `api_key` value.
            values (Mapping[str, Any]): Currently parsed attributes. `instance_name` is checked.

        Raises:
            ValueError: If `api_key` is undefined when `instance_name` is also undefined.

        Returns:
            Validated `api_key` value
        """
        if not values.get("instance_name", None) and not value:
            raise ValueError("required if 'instance_name' is undefined")
        return value

    @validator(
        "source_quality_profiles",
        "source_language_profiles",
        "source_tags",
        each_item=True,
    )
    def validate_source_resource_ids(
        cls,
        value: Union[int, str],
        values: Dict[str, Any],
    ) -> Union[int, str]:
        """
        Validate that all resource references are IDs (integers) if `instance_name` is undefined.

        Args:
            value (Union[int, str]): Resource reference (ID or name).
            values (Mapping[str, Any]): Currently parsed attributes. `instance_name` is checked.

        Raises:
            ValueError: If the resource reference is a name and `instance_name` is undefined.

        Returns:
            Validated resource reference
        """
        if not values.get("instance_name", None) and not isinstance(value, int):
            raise ValueError(
                "values must be IDs (not names) if 'instance_name' is undefined",
            )
        return value

    @classmethod
    def _encode_source_resources(
        cls,
        instance_name: Optional[str],
        resources: Iterable[Union[str, int]],
        resource_type: str,
        name_key: str = "name",
    ) -> List[int]:
        """
        Encode a collection of resource IDs/names into a list of resource IDs
        from the target Sonarr instance.

        Args:
            instance_name (Optional[str]): Target Sonarr instance to get resource IDs from.
            resources (Iterable[Union[str, int]]): Resource names/IDs to encode.
            resource_type (str): Type of resource to encode into IDs.
            name_key (str, optional): Key for the name of the resource. Defaults to `name`.

        Returns:
            List of resource IDs for the target Sonarr instance
        """
        resource_ids: Set[int] = set()
        if not instance_name:
            for resource in resources:
                if not isinstance(resource, int):
                    raise RuntimeError(
                        f"{resource_type} reference should be of type int here: {resource}",
                    )
                resource_ids.add(resource)
            return sorted(resource_ids)
        source_resource_ids: Optional[Dict[str, int]] = None
        for i, resource in enumerate(resources):
            if isinstance(resource, int):
                resource_ids.add(resource)
            else:
                if source_resource_ids is None:
                    source_resource_ids = {
                        p[name_key]: p["id"]
                        for p in cls._get_resources(
                            instance_name,
                            resource_type,
                        )
                    }
                try:
                    resource_ids.add(source_resource_ids[resource])
                except KeyError as err:
                    # If this is a dry-run, then a non-existent resource name was passed from
                    # the instance reference resolver.
                    # Generate a fake ID based on the maximum existing resource ID
                    # and the index of the resource name being currently processed, and
                    # add it to the result.
                    # (Since this is a dry-run, it's fine that it's fake.)
                    if state.dry_run:
                        max_id = max(source_resource_ids.values()) if resource_ids else 1
                        resource_ids.add(max_id + i)
                    # If this is **not** a dry-run, then a non-existent resource name
                    # somehow made it past all the validation being done.
                    # Raise a runtime error and report back to the user.
                    else:
                        raise RuntimeError(
                            f"Unable to find ID for {resource_type} '{err.args[0]}'",
                        ) from None
        return sorted(resource_ids)

    def _resolve_from_local(
        self,
        name: str,
        local: Self,
        ignore_nonexistent_ids: bool = False,
    ) -> Self:
        instance_name = local.instance_name
        if not instance_name:
            return self
        api_key = self._get_secrets(instance_name).api_key
        source_quality_profiles = self._resolve_resources(
            name=name,
            instance_name=instance_name,
            source_resources=self.source_quality_profiles,
            resource_type_int="quality_profile",
            resource_type_ext="qualityprofile",
            resource_description="quality profile",
            ignore_nonexistent_ids=ignore_nonexistent_ids,
        )
        source_language_profiles = self._resolve_resources(
            name=name,
            instance_name=instance_name,
            source_resources=self.source_language_profiles,
            resource_type_int="language_profile",
            resource_type_ext="languageprofile",
            resource_description="language profile",
            ignore_nonexistent_ids=ignore_nonexistent_ids,
        )
        source_tags = self._resolve_resources(
            name=name,
            instance_name=instance_name,
            source_resources=self.source_tags,
            resource_type_int="tag",
            resource_type_ext="tag",
            resource_description="tag",
            ignore_nonexistent_ids=ignore_nonexistent_ids,
            name_key="label",
        )
        return self.copy(
            update={
                "instance_name": instance_name,
                "api_key": api_key,
                "source_quality_profiles": source_quality_profiles,
                "source_language_profiles": source_language_profiles,
                "source_tags": source_tags,
            },
        )

    def _resolve_resources(
        self,
        name: str,
        instance_name: str,
        source_resources: Iterable[Union[int, str]],
        resource_type_int: str,
        resource_type_ext: str,
        resource_description: str,
        ignore_nonexistent_ids: bool,
        name_key: str = "name",
    ) -> Set[Union[int, str]]:
        """
        Resolve target Sonarr instance resource IDs/names into resource names.

        If `ignore_nonexistent_ids` is `True` and a resource ID was not found
        on the Sonarr instance, it is returned as-is.
        This will prompt Buildarr to remove the offending ID from Sonarr,
        so a warning is output to the logs to notify the user.

        Args:
            name (str): Name associated with this import list.
            instance_name (str): Target Sonarr instance name in Buildarr.
            source_resources (Iterable[Union[int, str]]):
            resource_type (str): Type of resource to resolve IDs for names.
            resource_description (str): Description of the resource type for logging.
            ignore_nonexistent_ids (bool): If `True`, remove non-existent IDs from the remote.
            name_key (str, optional): _description_. Defaults to "name".

        Raises:
            ValueError: If a non-existent ID was found and `ignore_nonexistent_ids` is `False`.
            ValueError: If a resource name was not found on the target Sonarr instance.

        Returns:
            List of resolved source resource names (and invalid IDs)
        """
        resolved_source_resources: Set[Union[int, str]] = set()
        if not source_resources:
            return resolved_source_resources
        remote_resources = self._get_resources(instance_name, resource_type_ext)
        resource_ids = {r[name_key]: r["id"] for r in remote_resources}
        resource_names = {r["id"]: r[name_key] for r in remote_resources}
        for resource in source_resources:
            if isinstance(resource, int):
                try:
                    resolved_source_resources.add(resource_names[resource])
                except KeyError:
                    if ignore_nonexistent_ids:
                        logger.warning(
                            (
                                "Source %s ID %i referenced by remote Sonarr instance "
                                "not found on target instance '%s', removing"
                            ),
                            resource_description,
                            resource,
                            instance_name,
                        )
                        resolved_source_resources.add(resource)
                    else:
                        raise ValueError(
                            f"Source {resource_description} ID {resource} "
                            f"not found on target instance '{instance_name}",
                        ) from None
            elif resource in resource_ids:
                resolved_source_resources.add(resource)
            else:
                # If dry-run mode is enabled, then there is a very good chance
                # the resource does not yet exist on the target Sonarr instance.
                # Since the instance is also managed by Buildarr, go into its configuration
                # and check if the resource name is defined.
                #
                # If the resource is defined, continue as normal, and a fake ID will be assigned
                # when encoding the request, so the dry run will complete as expected.
                #
                # If the resource is **not** defined, raise an error as it would
                # in the non-dry-run case, but set `available_resources` so that
                # the error message will be accurate.
                if state.dry_run:
                    target_instance_config: SonarrInstanceConfig = state.instance_configs["sonarr"][
                        instance_name
                    ]  # type: ignore[assignment]
                    settings = target_instance_config.settings
                    definitions: Collection[str] = getattr(
                        settings.profiles if resource_type_int.endswith("_profile") else settings,
                        f"{resource_type_int}s",
                    ).definitions
                    if resource in definitions:
                        resolved_source_resources.add(resource)
                        continue
                    available_resources = list(definitions)
                # If this is not a dry-run, then either the user has not defined the
                # resource on the target instance, or for some reason it wasn't created.
                # Either way, raise an error.
                else:
                    available_resources = list(resource_ids.keys())
                _resources_str = (
                    ", ".join(repr(p) for p in available_resources)
                    if available_resources
                    else "(none)"
                )
                error_message = (
                    f"Source {resource_description} '{resource}' "
                    f"not found on target Sonarr instance '{instance_name}' "
                    f"in import list '{name}' "
                    f"(available {resource_description}s: {_resources_str})"
                )
                raise ValueError(error_message)
        return resolved_source_resources

    class Config(SonarrConfigBase.Config):
        # Ensure in-place assignments of attributes are always validated,
        # since this class performs such modifications in certain cases.
        validate_assignment = True


class PlexWatchlistImportList(PlexImportList):
    """
    Import items from a Plex watchlist.
    """

    type: Literal["plex-watchlist"] = "plex-watchlist"
    """
    Type value associated with this kind of import list.
    """

    access_token: Password
    """
    Plex authentication token.

    If unsure on where to find this token,
    [follow this guide from Plex.tv][PATH].
    [PATH]: https://support.plex.tv/articles/204059436-finding-an-authentication-token-x-plex-token
    """

    _implementation_name: str = "Plex Watchlist"
    _implementation: str = "PlexImport"
    _config_contract: str = "PlexListSettings"
    _remote_map: List[RemoteMapEntry] = [("access_token", "accessToken", {"is_field": True})]


class TraktListImportList(TraktImportList):
    """
    Import an arbitrary list from Trakt into Sonarr.
    """

    type: Literal["trakt-list"] = "trakt-list"
    """
    Type value associated with this kind of import list.
    """

    list_name: NonEmptyStr
    """
    Name of the list to import.

    The list must be public, or you must have access to the list.
    """

    _implementation_name: str = "Trakt List"
    _implementation: str = "TraktListImport"
    _config_contract: str = "TraktListSettings"
    _remote_map: List[RemoteMapEntry] = [("list_name", "listName", {"is_field": True})]


class TraktPopularlistImportList(TraktImportList):
    """
    Import media according to popularity-based lists on Trakt.
    """

    type: Literal["trakt-popularlist"] = "trakt-popularlist"
    """
    Type value associated with this kind of import list.
    """

    list_type: TraktPopularListType = TraktPopularListType.popular
    """
    Popularity-based list to import.

    Values:

    * `trending`
    * `popular`
    * `anticipated`
    * `top_watched_by_week`
    * `top_watched_by_month`
    * `top_watched_by_year`
    * `top_watched_by_alltime`
    * `recommended_by_week`
    * `recommended_by_month`
    * `recommended_by_year`
    * `recommended_by_alltime`
    """

    _implementation_name: str = "Trakt Popular List"
    _implementation: str = "TraktPopularImport"
    _config_contract: str = "TraktPopularSettings"
    _remote_map: List[RemoteMapEntry] = [("list_type", "traktListType", {"is_field": True})]


class TraktUserImportList(TraktImportList):
    """
    Import a user-level list from Trakt.
    """

    type: Literal["trakt-user"] = "trakt-user"
    """
    Type value associated with this kind of import list.
    """

    list_type: TraktUserListType = TraktUserListType.user_watch_list
    """
    User list type to import.

    Values:

    * `user_watch_list`
    * `user_watched_list`
    * `user_collection_list`
    """

    _implementation_name: str = "Trakt User"
    _implementation: str = "TraktUserImport"
    _config_contract: str = "TraktUserSettings"
    _remote_map: List[RemoteMapEntry] = [("list_type", "traktListType", {"is_field": True})]


IMPORTLIST_TYPES: Tuple[Type[ImportList], ...] = (
    SonarrImportList,
    PlexWatchlistImportList,
    TraktListImportList,
    TraktPopularlistImportList,
    TraktUserImportList,
)

IMPORTLIST_TYPE_MAP: Dict[str, Type[ImportList]] = {
    importlist_type._implementation: importlist_type for importlist_type in IMPORTLIST_TYPES
}

ImportListType = Union[
    SonarrImportList,
    PlexWatchlistImportList,
    TraktListImportList,
    TraktPopularlistImportList,
    TraktUserImportList,
]


class SonarrImportListsSettingsConfig(SonarrConfigBase):
    """
    Using import lists, Sonarr can monitor and import episodes from external sources.

    ```yaml
    sonarr:
      settings:
        import_lists:
          delete_unmanaged: False # Default is `false`
          delete_unmanaged_exclusions: true # Default is `false`
          definitions:
            Plex: # Name of import list definition
              type: "plex-watchlist" # Type of import list to use
              # Attributes common to all watch list types
              enable_automatic_add: true
              monitor: "all-episodes"
              series_type: "standard"
              season_folder: true
              tags:
                - "example"
              # Plex-specific attributes
              access_token: "..."
            # Add more import lists here.
          exclusions:
            72662: "Teletubbies" # TVDB ID is key, set an artibrary title as value
    ```

    Media can be queued on the source, and Sonarr will automatically import them,
    look for suitable releases, and download them.

    Media that you don't want to import can be ignored using the `exclusions`
    attribute.
    """

    delete_unmanaged: bool = False
    """
    Automatically delete import lists not defined in Buildarr.
    """

    delete_unmanaged_exclusions: bool = False
    """
    Automatically delete import list excusions not defined in Buildarr.
    """

    definitions: Dict[str, Annotated[ImportListType, Field(discriminator="type")]] = {}
    """
    Import list definitions go here.
    """

    exclusions: Dict[PositiveInt, NonEmptyStr] = {}
    """
    Dictionary of TV series that should be excluded from being imported.

    The key is the TVDB ID of the series to exclude, the value is
    a title to give the series in the Sonarr UI.
    """

    @classmethod
    def from_remote(cls, secrets: SonarrSecrets) -> Self:
        importlists = api_get(secrets, "/api/v3/importlist")
        quality_profile_ids: Dict[str, int] = (
            {pro["name"]: pro["id"] for pro in api_get(secrets, "/api/v3/qualityprofile")}
            if any(importlist["qualityProfileId"] for importlist in importlists)
            else {}
        )
        language_profile_ids: Dict[str, int] = (
            {pro["name"]: pro["id"] for pro in api_get(secrets, "/api/v3/languageprofile")}
            if any(importlist["languageProfileId"] for importlist in importlists)
            else {}
        )
        tag_ids: Dict[str, int] = (
            {tag["label"]: tag["id"] for tag in api_get(secrets, "/api/v3/tag")}
            if any(importlist["tags"] for importlist in importlists)
            else {}
        )
        return cls(
            definitions={
                importlist["name"]: IMPORTLIST_TYPE_MAP[importlist["implementation"]]._from_remote(
                    quality_profile_ids=quality_profile_ids,
                    language_profile_ids=language_profile_ids,
                    tag_ids=tag_ids,
                    remote_attrs=importlist,
                )
                for importlist in importlists
            },
        )

    def update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> bool:
        # Flag for whether or not the import list configuration was updated or not.
        changed = False
        # Get required resource ID references from the remote Sonarr instance.
        importlist_ids: Dict[str, int] = {
            importlist_json["name"]: importlist_json["id"]
            for importlist_json in api_get(secrets, "/api/v3/importlist")
        }
        quality_profile_ids: Dict[str, int] = {
            pro["name"]: pro["id"] for pro in api_get(secrets, "/api/v3/qualityprofile")
        }
        language_profile_ids: Dict[str, int] = {
            pro["name"]: pro["id"] for pro in api_get(secrets, "/api/v3/languageprofile")
        }
        tag_ids: Dict[str, int] = {
            tag["label"]: tag["id"] for tag in api_get(secrets, "/api/v3/tag")
        }
        # Evaluate locally defined import lists against the currently active ones
        # on the remote instance.
        for importlist_name, importlist in self.definitions.items():
            importlist = importlist._resolve(importlist_name)  # noqa: PLW2901
            importlist_tree = f"{tree}.definitions[{repr(importlist_name)}]"
            # If a locally defined import list does not exist on the remote, create it.
            if importlist_name not in remote.definitions:
                importlist._create_remote(
                    tree=importlist_tree,
                    secrets=secrets,
                    quality_profile_ids=quality_profile_ids,
                    language_profile_ids=language_profile_ids,
                    tag_ids=tag_ids,
                    importlist_name=importlist_name,
                )
                changed = True
            # Since there is an import list with the same name on the remote,
            # update it in-place.
            elif importlist._update_remote(
                tree=importlist_tree,
                secrets=secrets,
                remote=remote.definitions[importlist_name]._resolve_from_local(
                    name=importlist_name,
                    local=importlist,  # type: ignore[arg-type]
                    ignore_nonexistent_ids=True,
                ),
                quality_profile_ids=quality_profile_ids,
                language_profile_ids=language_profile_ids,
                tag_ids=tag_ids,
                importlist_id=importlist_ids[importlist_name],
                importlist_name=importlist_name,
            ):
                changed = True
        # Find import list definitions on the remote that aren't configured locally.
        # If `delete_unmanaged` is `True`, delete them. If not, just log them as unmanaged.
        for importlist_name, importlist in remote.definitions.items():
            if importlist_name not in self.definitions:
                importlist_tree = f"{tree}.definitions[{repr(importlist_name)}]"
                if self.delete_unmanaged:
                    importlist._delete_remote(
                        tree=importlist_tree,
                        secrets=secrets,
                        importlist_id=importlist_ids[importlist_name],
                    )
                    changed = True
                else:
                    logger.debug("%s: (...) (unmanaged)", importlist_tree)
        # We're done!
        return changed



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin connect settings configuration.
"""


from __future__ import annotations

from datetime import datetime
from logging import getLogger
from typing import Any, Dict, List, Literal, Mapping, Optional, Set, Tuple, Type, Union

from buildarr.config import RemoteMapEntry
from buildarr.types import BaseEnum, BaseIntEnum, NonEmptyStr, Password, Port
from pydantic import AnyHttpUrl, ConstrainedInt, Field, NameEmail, SecretStr
from typing_extensions import Annotated, Self

from ..api import api_delete, api_get, api_post, api_put
from ..secrets import SonarrSecrets
from .types import SonarrConfigBase, TraktAuthUser
from .util import trakt_expires_encoder

logger = getLogger(__name__)


class OnGrabField(BaseIntEnum):
    """
    Values for `on_grab_fields` for the Discord connection.
    """

    overview = 0
    rating = 1
    genres = 2
    quality = 3
    group = 4
    size = 5
    links = 6
    release = 7
    poster = 8
    fanart = 9


class OnImportField(BaseIntEnum):
    """
    Values for `on_import_fields` for the Discord connection.
    """

    overview = 0
    rating = 1
    genres = 2
    quality = 3
    codecs = 4
    group = 5
    size = 6
    languages = 7
    subtitles = 8
    links = 9
    release = 10
    poster = 11
    fanart = 12


class GotifyPriority(BaseEnum):
    """
    Gotify notification priority.
    """

    min = 0
    low = 2
    normal = 5
    high = 8


class JoinPriority(BaseEnum):
    """
    Join notification priority.
    """

    silent = -2
    quiet = -1
    normal = 0
    high = 1
    emergency = 2


class ProwlPriority(BaseEnum):
    """
    Prowl notification priority.
    """

    verylow = -2
    low = -1
    normal = 0
    high = 1
    emergency = 2


class PushoverPriority(BaseEnum):
    """
    Pushover notification priority.
    """

    silent = -2
    quiet = -1
    normal = 0
    high = 1
    emergency = 2


class PushoverRetry(ConstrainedInt):
    """
    Constrained integer type to enforce Pushover retry field limits.
    """

    ge = 30


class WebhookMethod(BaseEnum):
    """
    HTTP method to use on a webhook connection.
    """

    POST = 1
    PUT = 2


class NotificationTriggers(SonarrConfigBase):
    """
    Connections are configured using the following syntax.

    ```yaml
    sonarr:
      settings:
        connect:
          delete_unmanaged: false # Optional
          definitions:
            Email: # Name of connection in Sonarr.
              type: "email" # Required
              notification_triggers: # When to send notifications.
                on_grab: true
                on_import: true
                on_upgrade: true
                on_rename: false # Not supported by email notifications.
                on_series_delete: true
                on_episode_file_delete: true
                on_episode_file_delete_for_upgrade: true
                on_health_issue: true
                include_health_warnings: false # Do not send on just warnings.
                on_application_update: true
              tags: # Tags can also be assigned to connections.
                - "example"
              # Connection-specific parameters.
              server: "smtp.example.com"
              port: 465
              use_encryption: true
              username: "sonarr"
              password: "fake-password"
              from_address: "sonarr@example.com"
              recipient_addresses:
                - "admin@example.com"
            # Add additional connections here.
    ```

    A `type` attribute must be defined so Buildarr knows what type of connection to make.
    Each connection has a unique value for `type` documented below.

    The triggers enabled on a connection are defined under `notification_triggers`.
    Tags can be assigned to connections, to only allow notifications relating
    to media under those tags.

    The `delete_unmanaged` flag on the outer `connect` block can be set
    to remove connections not defined in Buildarr.
    Take care when using this option, as it can remove connections
    automatically managed by other applications.

    The following notification triggers can be enabled.
    Some connection types only allow a subset of these to be enabled,
    check the documentation the specific connection type for more information.
    """

    on_grab: bool = False
    """
    Be notified when episodes are available for download and has been sent to a download client.
    """

    on_import: bool = False
    """
    Be notified when episodes are successfully imported. (Formerly known as On Download)
    """

    on_upgrade: bool = False
    """
    Be notified when episodes are upgraded to a better quality.
    """

    on_rename: bool = False
    """
    Be notified when episodes are renamed.
    """

    on_series_delete: bool = False
    """
    Be notified when series are deleted.
    """

    on_episode_file_delete: bool = False
    """
    Be notified when episodes files are deleted.
    """

    on_episode_file_delete_for_upgrade: bool = False
    """
    Be notified when episode files are deleted for upgrades.
    """

    on_health_issue: bool = False
    """
    Be notified on health check failures.
    """

    include_health_warnings: bool = False
    """
    Be notified on health warnings in addition to errors.

    Requires `on_health_issue` to be enabled to have any effect.
    """

    on_application_update: bool = False
    """
    Be notified when Sonarr gets updated to a new version.
    """

    _remote_map: List[RemoteMapEntry] = [
        ("on_grab", "onGrab", {}),
        ("on_import", "onDownload", {}),
        ("on_upgrade", "onUpgrade", {}),
        ("on_rename", "onRename", {}),
        ("on_series_delete", "onSeriesDelete", {}),
        ("on_episode_file_delete", "onEpisodeFileDelete", {}),
        ("on_episode_file_delete_for_upgrade", "onEpisodeFileDeleteForUpgrade", {}),
        ("on_health_issue", "onHealthIssue", {}),
        ("include_health_warnings", "includeHealthWarnings", {}),
        ("on_application_update", "onApplicationUpdate", {}),
    ]


class Connection(SonarrConfigBase):
    """
    Base class for a Sonarr connection.
    """

    notification_triggers: NotificationTriggers = NotificationTriggers()
    """
    Notification triggers to enable on this connection.
    """

    tags: List[NonEmptyStr] = []
    """
    Sonarr tags to associate this connection with.
    """

    _implementation_name: str
    _implementation: str
    _config_contract: str
    _remote_map: List[RemoteMapEntry]

    @classmethod
    def _get_base_remote_map(
        cls,
        tag_ids: Mapping[str, int],
    ) -> List[RemoteMapEntry]:
        return [
            (
                "tags",
                "tags",
                {
                    "decoder": lambda v: set(
                        (tag for tag, tag_id in tag_ids.items() if tag_id in v),
                    ),
                    "encoder": lambda v: sorted(tag_ids[tag] for tag in v),
                },
            ),
        ]

    @classmethod
    def _from_remote(cls, tag_ids: Mapping[str, int], remote_attrs: Mapping[str, Any]) -> Self:
        return cls(
            notification_triggers=NotificationTriggers(
                **NotificationTriggers.get_local_attrs(
                    remote_map=NotificationTriggers._remote_map,
                    remote_attrs=remote_attrs,
                ),
            ),
            **cls.get_local_attrs(
                remote_map=cls._get_base_remote_map(tag_ids) + cls._remote_map,
                remote_attrs=remote_attrs,
            ),
        )

    def _create_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        tag_ids: Mapping[str, int],
        connection_name: str,
    ) -> None:
        api_post(
            secrets,
            "/api/v3/notification",
            {
                "name": connection_name,
                "implementation": self._implementation,
                "implementationName": self._implementation_name,
                "configContract": self._config_contract,
                **self.notification_triggers.get_create_remote_attrs(
                    tree=f"{tree}.notification_triggers",
                    remote_map=self.notification_triggers._remote_map,
                ),
                **self.get_create_remote_attrs(
                    tree=tree,
                    remote_map=self._get_base_remote_map(tag_ids) + self._remote_map,
                ),
            },
        )

    def _update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        tag_ids: Mapping[str, int],
        connection_id: int,
        connection_name: str,
    ) -> bool:
        (
            triggers_updated,
            triggers_remote_attrs,
        ) = self.notification_triggers.get_update_remote_attrs(
            tree=tree,
            remote=remote.notification_triggers,
            remote_map=self.notification_triggers._remote_map,
            # TODO: check if check_unmanaged and/or set_unchanged are required (probably are)
        )
        base_updated, base_remote_attrs = self.get_update_remote_attrs(
            tree=tree,
            remote=remote,
            remote_map=self._get_base_remote_map(tag_ids) + self._remote_map,
            # TODO: check if check_unmanaged and/or set_unchanged are required (probably are)
        )
        if triggers_updated or base_updated:
            api_put(
                secrets,
                f"/api/v3/notification/{connection_id}",
                {
                    "id": connection_id,
                    "name": connection_name,
                    "implementation": self._implementation,
                    "implementationName": self._implementation_name,
                    "configContract": self._config_contract,
                    **triggers_remote_attrs,
                    **base_remote_attrs,
                },
            )
            return True
        return False

    def _delete_remote(self, tree: str, secrets: SonarrSecrets, connection_id: int) -> None:
        logger.info("%s: (...) -> (deleted)", tree)
        api_delete(secrets, f"/api/v3/notification/{connection_id}")


class BoxcarConnection(Connection):
    """
    Receive media update and health alert push notifications via Boxcar.

    Supported notification triggers: All except `on_rename`
    """

    type: Literal["boxcar"] = "boxcar"
    """
    Type value associated with this kind of connection.
    """

    access_token: Password
    """
    Access token for authenticating with Boxcar.
    """

    _implementation_name: str = "Boxcar"
    _implementation: str = "Boxcar"
    _config_contract: str = "BoxcarSettings"
    _remote_map: List[RemoteMapEntry] = [("access_token", "token", {"is_field": True})]


class CustomscriptConnection(Connection):
    """
    Execute a local script on the Sonarr instance when events occur.

    Supported notification triggers: All
    """

    type: Literal["customscript"] = "customscript"
    """
    Type value associated with this kind of connection.
    """

    path: NonEmptyStr
    """
    Path of the script to execute.
    """

    _implementation_name: str = "Custom Script"
    _implementation: str = "CustomScript"
    _config_contract: str = "CustomScriptSettings"
    _remote_map: List[RemoteMapEntry] = [("path", "path", {"is_field": True})]


class DiscordConnection(Connection):
    """
    Send media update and health alert messages to a Discord server.

    Supported notification triggers: All
    """

    type: Literal["discord"] = "discord"
    """
    Type value associated with this kind of connection.
    """

    webhook_url: AnyHttpUrl
    """
    Discord server webhook URL.
    """

    username: Optional[str] = None
    """
    The username to post as.

    If unset, blank or set to `None`, use the default username set to the webhook URL.
    """

    avatar: Optional[str] = None
    """
    Change the avatar that is used for messages from this connection.

    If unset, blank or set to `None`, use the default avatar for the user.
    """

    # Name override, None -> use machine_name
    host: Optional[str] = None
    """
    Override the host name that shows for this notification.

    If unset, blank or set to `None`, use the machine name.
    """

    on_grab_fields: Set[OnGrabField] = {
        OnGrabField.overview,
        OnGrabField.rating,
        OnGrabField.genres,
        OnGrabField.quality,
        OnGrabField.size,
        OnGrabField.links,
        OnGrabField.release,
        OnGrabField.poster,
        OnGrabField.fanart,
    }
    """
    Set the fields that are passed in for this 'on grab' notification.
    By default, all fields are passed in.

    Values:

    * `overview`
    * `rating`
    * `genres`
    * `quality`
    * `group`
    * `size`
    * `links`
    * `release`
    * `poster`
    * `fanart`

    Example:

    ```yaml
    ...
      connect:
        definitions:
          Discord:
            type: "discord"
            webhook_url: "https://..."
            on_grab_fields:
              - "overview"
              - "quality"
              - "release"
    ```
    """

    on_import_fields: Set[OnImportField] = {
        OnImportField.overview,
        OnImportField.rating,
        OnImportField.genres,
        OnImportField.quality,
        OnImportField.codecs,
        OnImportField.group,
        OnImportField.size,
        OnImportField.languages,
        OnImportField.subtitles,
        OnImportField.links,
        OnImportField.release,
        OnImportField.poster,
        OnImportField.fanart,
    }
    """
    Set the fields that are passed in for this 'on import' notification.
    By default, all fields are passed in.

    Values:

    * `overview`
    * `rating`
    * `genres`
    * `quality`
    * `codecs`
    * `group`
    * `size`
    * `languages`
    * `subtitles`
    * `links`
    * `release`
    * `poster`
    * `fanart`

    Example:

    ```yaml
    ...
      connect:
        definitions:
          Discord:
            type: "discord"
            webhook_url: "https://..."
            on_import_fields:
              - "overview"
              - "quality"
              - "release"
    ```
    """

    _implementation_name: str = "Discord"
    _implementation: str = "Discord"
    _config_contract: str = "DiscordSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("webhook_url", "webHookUrl", {"is_field": True}),
        (
            "username",
            "username",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        (
            "avatar",
            "avatar",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        (
            "host",
            "host",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        (
            "on_grab_fields",
            "grabFields",
            {"is_field": True, "encoder": lambda v: sorted(f.value for f in v)},
        ),
        (
            "on_import_fields",
            "importFields",
            {"is_field": True, "encoder": lambda v: sorted(f.value for f in v)},
        ),
    ]


class EmailConnection(Connection):
    """
    Send media update and health alert messages to an email address.

    Supported notification triggers: All except `on_rename`
    """

    type: Literal["email"] = "email"
    """
    Type value associated with this kind of connection.
    """

    server: NonEmptyStr
    """
    Hostname or IP address of the SMTP server to send outbound mail to.
    """

    port: Port = 587  # type: ignore[assignment]
    """
    The port number on the SMTP server to use to submit mail.

    The default is to use STARTTLS on the standard SMTP submission port.
    """

    use_encryption: bool = True
    """
    Whether or not to use encryption when sending mail to the SMTP server.

    If the port number is set to 465, SMTPS (implicit TLS) will be used.
    Any other port number will result in STARTTLS being used.

    The default is to enable encryption.
    """

    username: NonEmptyStr
    """
    SMTP username of the account to send the mail from.
    """

    password: Password
    """
    SMTP password of the account to send the mail from.
    """

    from_address: NameEmail
    """
    Email address to send the mail as.

    RFC-5322 formatted mailbox addresses are also supported,
    e.g. `Sonarr Notifications <sonarr@example.com>`.
    """

    recipient_addresses: Annotated[List[NameEmail], Field(min_items=1, unique_items=True)]
    """
    List of email addresses to directly address the mail to.

    At least one address must be provided.
    """

    cc_addresses: Annotated[List[NameEmail], Field(unique_items=True)] = []
    """
    Optional list of email addresses to copy (CC) the mail to.
    """

    bcc_addresses: Annotated[List[NameEmail], Field(unique_items=True)] = []
    """
    Optional list of email addresses to blind copy (BCC) the mail to.
    """

    _implementation_name: str = "Email"
    _implementation: str = "Email"
    _config_contract: str = "EmailSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("server", "server", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_encryption", "requireEncryption", {"is_field": True}),
        ("username", "username", {"is_field": True}),
        ("password", "password", {"is_field": True}),
        ("from_address", "from", {"is_field": True}),
        ("recipient_addresses", "to", {"is_field": True}),
        ("cc_addresses", "cc", {"is_field": True}),
        ("bcc_addresses", "bcc", {"is_field": True}),
    ]


class EmbyConnection(Connection):
    """
    Send media update and health alert notifications to an Emby server.

    Supported notification triggers: All
    """

    type: Literal["emby"] = "emby"
    """
    Type value associated with this kind of connection.
    """

    host: NonEmptyStr
    """
    Emby server hostname or IP address.
    """

    port: Port = 8096  # type: ignore[assignment]
    """
    Emby server port.
    """

    use_ssl: bool = False
    """
    Use HTTPS to connect to Emby, instead of HTTP.
    """

    api_key: Password
    """
    API key to authenticate with Emby.
    """

    send_notifications: bool = False
    """
    Have MediaBrowser send notifications to configured providers.
    """

    update_library: bool = False
    """
    Update the Emby library on import, rename, or delete.
    """

    _implementation_name: str = "Emby"
    _implementation: str = "MediaBrowser"
    _config_contract: str = "MediaBrowserSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        ("api_key", "apiKey", {"is_field": True}),
        ("send_notifications", "notify", {"is_field": True}),
        ("update_library", "updateLibrary", {"is_field": True}),
    ]


class GotifyConnection(Connection):
    """
    Send media update and health alert push notifications via a Gotify server.

    Supported notification triggers: All except `on_rename`
    """

    type: Literal["gotify"] = "gotify"
    """
    Type value associated with this kind of connection.
    """

    server: AnyHttpUrl
    """
    Gotify server URL. (e.g. `http://gotify.example.com:1234`)
    """

    app_token: Password
    """
    App token to use to authenticate with Gotify.
    """

    priority: GotifyPriority = GotifyPriority.normal
    """
    Gotify notification priority.

    Values:

    * `min`
    * `low`
    * `normal`
    * `high`
    """

    _implementation: str = "Gotify"
    _implementation_name: str = "Gotify"
    _config_contract: str = "GotifySettings"
    _remote_map: List[RemoteMapEntry] = [
        ("server", "server", {"is_field": True}),
        ("app_token", "appToken", {"is_field": True}),
        ("priority", "priority", {"is_field": True}),
    ]


class JoinConnection(Connection):
    """
    Send media update and health alert push notifications via Join.

    Supported notification triggers: All except `on_rename`
    """

    type: Literal["join"] = "join"
    """
    Type value associated with this kind of connection.
    """

    api_key: Password
    """
    API key to use to authenticate with Join.
    """

    # Deprecated, only uncomment if absolutely required by Sonarr
    # device_ids: Set[int] = set()

    device_names: Set[NonEmptyStr] = set()
    """
    List of full or partial device names you'd like to send notifications to.

    If unset or empty, all devices will receive notifications.
    """

    priority: JoinPriority = JoinPriority.normal
    """
    Join push notification priority.

    Values:

    * `silent`
    * `quiet`
    * `normal`
    * `high`
    * `emergency`
    """

    _implementation: str = "Join"
    _implementation_name: str = "Join"
    _config_contract: str = "JoinSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("api_key", "apiKey", {"is_field": True}),
        ("device_names", "deviceNames", {"is_field": True}),
        (
            "device_names",
            "deviceNames",
            {
                "is_field": True,
                "decoder": lambda v: (
                    set(d.strip() for d in v.split(",")) if v and v.strip() else set()
                ),
                "encoder": lambda v: ",".join(sorted(v)) if v else "",
            },
        ),
        ("priority", "priority", {"is_field": True}),
    ]


class KodiConnection(Connection):
    """
    Send media update and health alert notifications to a Kodi (XBMC) instance.

    Supported notification triggers: All
    """

    type: Literal["kodi", "xbmc"] = "kodi"
    """
    Type values associated with this kind of connection. (Use either one)
    """

    host: NonEmptyStr
    """
    Kodi hostname or IP address.
    """

    port: Port = 8080  # type: ignore[assignment]
    """
    Kodi API port.
    """

    use_ssl: bool = False
    """
    Connect to Kodi over HTTPS instead of HTTP.
    """

    username: NonEmptyStr
    """
    Username to use to login to Kodi.
    """

    password: Password
    """
    Password to use to login to Kodi.
    """

    gui_notification: bool = False
    """
    Enable showing notifications from Sonarr in the Kodi GUI.
    """

    display_time: int = Field(5, ge=0)  # seconds
    """
    How long the notification will be displayed for (in seconds).
    """

    update_library: bool = False
    """
    Update the Kodi library on import/rename.
    """

    clean_library: bool = False
    """
    Clean the Kodi library after update.
    """

    always_update: bool = False
    """
    Update the Kodi library even when a video is playing.
    """

    _implementation_name: str = "Kodi"
    _implementation: str = "Xbmc"
    _config_contract: str = "XbmcSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        ("username", "username", {"is_field": True}),
        ("password", "password", {"is_field": True}),
        ("gui_notification", "notify", {"is_field": True}),
        ("display_time", "displayTime", {"is_field": True}),
        ("update_library", "updateLibrary", {"is_field": True}),
        ("clean_library", "cleanLibrary", {"is_field": True}),
        ("always_update", "alwaysUpdate", {"is_field": True}),
    ]


class MailgunConnection(Connection):
    """
    Send media update and health alert emails via the Mailgun delivery service.

    Supported notification triggers: All except `on_rename`
    """

    type: Literal["mailgun"] = "mailgun"
    """
    Type value associated with this kind of connection.
    """

    api_key: Password
    """
    API key to use to authenticate with Mailgun.
    """

    use_eu_endpoint: bool = False
    """
    Send mail via the EU endpoint instead of the US one.
    """

    from_address: NameEmail
    """
    Email address to send the mail as.

    RFC-5322 formatted mailbox addresses are also supported,
    e.g. `Sonarr Notifications <sonarr@example.com>`.
    """

    sender_domain: NonEmptyStr
    """
    The domain from which the mail will be sent.
    """

    recipient_addresses: Annotated[List[NameEmail], Field(min_items=1, unique_items=True)]
    """
    The recipient email addresses of the notification mail.

    At least one recipient address is required.
    """

    _implementation: str = "Mailgun"
    _implementation_name: str = "MailGun"
    _config_contract: str = "MailgunSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("api_key", "apiKey", {"is_field": True}),
        ("use_eu_endpoint", "useEuEndpoint", {"is_field": True}),
        ("from_address", "from", {"is_field": True}),
        ("sender_domain", "senderDomain", {"is_field": True}),
        ("recipient_addresses", "recipients", {"is_field": True}),
    ]


class PlexHomeTheaterConnection(Connection):
    """
    Send media update notifications to a Plex Home Theater instance.

    Supported notification triggers: `on_grab`, `on_import`, `on_upgrade` **only**
    """

    type: Literal["plex-home-theater"] = "plex-home-theater"
    """
    Type value associated with this kind of connection.
    """

    host: NonEmptyStr
    """
    Plex Home Theater hostname or IP address.
    """

    port: Port = 3005  # type: ignore[assignment]
    """
    Plex Home Theater API port.
    """

    use_ssl: bool = False
    """
    Connect to Plex Home Theater over HTTPS instead of HTTP.
    """

    username: NonEmptyStr
    """
    Username to use to login to Plex Home Theater.
    """

    password: Password
    """
    Password to use to login to Plex Home Theater.
    """

    gui_notification: bool = False
    """
    Enable showing notifications from Sonarr in the Plex Home Theater GUI.
    """

    display_time: int = Field(5, ge=0)  # seconds
    """
    How long the notification will be displayed for (in seconds).
    """

    update_library: bool = False
    """
    Update the Plex Home Theater library on import/rename.
    """

    clean_library: bool = False
    """
    Clean the Plex Home Theater library after update.
    """

    always_update: bool = False
    """
    Update the Plex Home THeater library even when a video is playing.
    """

    _implementation_name: str = "Plex Home Theater"
    _implementation: str = "PlexHomeTheater"
    _config_contract: str = "PlexHomeTheaterSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        ("username", "username", {"is_field": True}),
        ("password", "password", {"is_field": True}),
        ("gui_notification", "notify", {"is_field": True}),
        ("display_time", "displayTime", {"is_field": True}),
        ("update_library", "updateLibrary", {"is_field": True}),
        ("clean_library", "cleanLibrary", {"is_field": True}),
        ("always_update", "alwaysUpdate", {"is_field": True}),
    ]


class PlexMediaCenterConnection(Connection):
    """
    Send media update notifications to a Plex Media Center instance.

    Supported notification triggers: `on_grab`, `on_import`, `on_upgrade` **only**
    """

    type: Literal["plex-media-center"] = "plex-media-center"
    """
    Type value associated with this kind of connection.
    """

    host: NonEmptyStr
    """
    Plex Media Center hostname or IP address.
    """

    port: Port = 3000  # type: ignore[assignment]
    """
    Plex Media Center API port.
    """

    username: NonEmptyStr
    """
    Username to use to login to Plex Media Center.
    """

    password: Password
    """
    Password to use to login to Plex Media Center.
    """

    _implementation_name: str = "Plex Media Center"
    _implementation: str = "PlexClient"
    _config_contract: str = "PlexClientSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("username", "username", {"is_field": True}),
        ("password", "password", {"is_field": True}),
    ]


class PlexMediaServerConnection(Connection):
    """
    Send media update notifications to a Plex Media Server instance.

    Supported notification triggers: All except `on_grab`, `on_health_issue`
    and `on_application_update`
    """

    type: Literal["plex-media-server"] = "plex-media-server"
    """
    Type value associated with this kind of connection.
    """

    host: NonEmptyStr
    """
    Plex Media Server hostname or IP address.
    """

    port: Port = 32400  # type: ignore[assignment]
    """
    Plex Media Server API port.
    """

    use_ssl: bool = False
    """
    Connect to Plex Media Server over HTTPS instead of HTTP.
    """

    auth_token: Password
    """
    Plex authentication token.

    If unsure on where to find this token,
    [follow this guide from Plex.tv][PATK].
    [PATK]: https://support.plex.tv/articles/204059436-finding-an-authentication-token-x-plex-token
    """

    # TODO: Determine how this should be handled.
    # I think as long as auth_token is specified, there should be no problems.
    # sign_in: Literal["startOAuth"] = "startOAuth"

    update_library: bool = True
    """
    Update the Plex Media Server library on import, rename or delete.
    """

    _implementation_name: str = "Plex Media Server"
    _implementation: str = "PlexServer"
    _config_contract: str = "PlexServerSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        ("auth_token", "authToken", {"is_field": True}),
        ("update_library", "updateLibrary", {"is_field": True}),
    ]


class ProwlConnection(Connection):
    """
    Send media update and health alert push notifications to a Prowl client.

    Supported notification triggers: All except `on_rename`
    """

    type: Literal["prowl"] = "prowl"
    """
    Type value associated with this kind of connection.
    """

    api_key: Password
    """
    API key to use when authenticating with Prowl.
    """

    priority: ProwlPriority = ProwlPriority.normal
    """
    Prowl push notification priority.

    Values:

    * `verylow`
    * `low`
    * `normal`
    * `high`
    * `emergency`
    """

    _implementation: str = "Prowl"
    _implementation_name: str = "Prowl"
    _config_contract: str = "ProwlSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("api_key", "apiKey", {"is_field": True}),
        ("priority", "priority", {"is_field": True}),
    ]


class PushbulletConnection(Connection):
    """
    Send media update and health alert push notifications to 1 or more Pushbullet devices.

    Supported notification triggers: All except `on_rename`
    """

    type: Literal["pushbullet"] = "pushbullet"
    """
    Type value associated with this kind of connection.
    """

    api_key: Password
    """
    API key to use when authenticating with Pushbullet.
    """

    device_ids: List[NonEmptyStr] = []
    """
    List of device IDs to send notifications to.

    If unset or empty, send to all devices.
    """

    channel_tags: List[NonEmptyStr] = []
    """
    List of Channel Tags to send notifications to.
    """

    sender_id: Optional[str] = None
    """
    The device ID to send notifications from
    (`device_iden` in the device's URL on [pushbullet.com](https://pushbullet.com)).

    Leave unset, blank or set to `None` to send from yourself.
    """

    _implementation: str = "Pushbullet"
    _implementation_name: str = "PushBullet"
    _config_contract: str = "PushBulletSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("api_key", "apiKey", {"is_field": True}),
        ("device_ids", "deviceIds", {"is_field": True}),
        ("channel_tags", "channelTags", {"is_field": True}),
        (
            "sender_id",
            "senderId",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
    ]


class PushoverConnection(Connection):
    """
    Send media update and health alert push notifications to 1 or more Pushover devices.

    Supported notification triggers: All except `on_rename`
    """

    type: Literal["pushover"] = "pushover"
    """
    Type value associated with this kind of connection.
    """

    user_key: Annotated[SecretStr, Field(min_length=30, max_length=30)]
    """
    User key to use to authenticate with your Pushover account.
    """

    api_key: Annotated[SecretStr, Field(min_length=30, max_length=30)]
    """
    API key assigned to Sonarr in Pushover.
    """

    devices: Set[NonEmptyStr] = set()
    """
    List of device names to send notifications to.

    If unset or empty, send to all devices.
    """

    priority: PushoverPriority = PushoverPriority.normal
    """
    Pushover push notification priority.

    Values:

    * `silent`
    * `quiet`
    * `normal`
    * `high`
    * `emergency`
    """

    retry: Union[Literal[0], PushoverRetry] = 0
    """
    Interval to retry emergency alerts, in seconds.

    Minimum 30 seconds. Set to 0 to disable retrying emergency alerts.
    """

    # TODO: Enforce "expire > retry if retry > 0" constraint
    expire: int = Field(0, ge=0, le=86400)
    """
    Threshold for retrying emergency alerts, in seconds.
    If `retry` is set, this should be set to a higher value.

    Maximum 86400 seconds (1 day).
    """

    sound: Optional[str] = None
    """
    Notification sound to use on devices.

    Leave unset, blank or set to `None` to use the default.
    """

    _implementation_name: str = "Pushover"
    _implementation: str = "Pushover"
    _config_contract: str = "PushoverSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("user_key", "userKey", {"is_field": True}),
        ("api_key", "apiKey", {"is_field": True}),
        ("devices", "devices", {"is_field": True, "encoder": lambda v: sorted(v)}),
        ("priority", "priority", {"is_field": True}),
        ("retry", "retry", {"is_field": True}),
        ("expire", "expire", {"is_field": True}),
        (
            "sound",
            "sound",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
    ]


class SendgridConnection(Connection):
    """
    Send media update and health alert emails via the SendGrid delivery service.

    Supported notification triggers: All except `on_rename`
    """

    type: Literal["sendgrid"] = "sendgrid"
    """
    Type value associated with this kind of connection.
    """

    api_key: Password
    """
    API key to use to authenticate with SendGrid.
    """

    from_address: NameEmail
    """
    Email address to send the mail as.

    RFC-5322 formatted mailbox addresses are also supported,
    e.g. `Sonarr Notifications <sonarr@example.com>`.
    """

    recipient_addresses: Annotated[List[NameEmail], Field(min_items=1, unique_items=True)]
    """
    The recipient email addresses of the notification mail.

    At least one recipient address is required.
    """

    _implementation: str = "SendGrid"
    _implementation_name: str = "SendGrid"
    _config_contract: str = "SendGridSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("api_key", "apiKey", {"is_field": True}),
        ("from_address", "from", {"is_field": True}),
        ("recipient_addresses", "recipients", {"is_field": True}),
    ]


class SlackConnection(Connection):
    """
    Send media update and health alert messages to a Slack channel.

    Supported notification triggers: All
    """

    type: Literal["slack"] = "slack"
    """
    Type value associated with this kind of connection.
    """

    webhook_url: AnyHttpUrl
    """
    Webhook URL for the Slack channel to send to.
    """

    username: NonEmptyStr
    """
    Username to post as.
    """

    icon: Optional[str] = None
    """
    The icon that is used for messages from this integration (emoji or URL).

    If unset, blank or set to `None`, use the default for the user.
    """

    channel: Optional[str] = None
    """
    If set, overrides the default channel in the webhook.
    """

    _implementation: str = "Slack"
    _implementation_name: str = "Slack"
    _config_contract: str = "SlackSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("webhook_url", "webHookUrl", {"is_field": True}),
        ("username", "username", {"is_field": True}),
        (
            "icon",
            "icon",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        (
            "channel",
            "channel",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
    ]


class SynologyIndexerConnection(Connection):
    """
    Send media update notifications to the local Synology Indexer.

    Supported notification triggers: All except `on_grab`, `on_health_issue`
    and `on_application_update`
    """

    type: Literal["synology-indexer"] = "synology-indexer"
    """
    Type value associated with this kind of connection.
    """

    update_library: bool = True
    """
    Update the library on media import/rename/delete.
    """

    _implementation_name: str = "Synology Indexer"
    _implementation: str = "SynologyIndexer"
    _config_contract: str = "SynologyIndexerSettings"
    _remote_map: List[RemoteMapEntry] = [("update_library", "updateLibrary", {"is_field": True})]


class TelegramConnection(Connection):
    """
    Send media update and health alert messages to a Telegram chat room.

    Supported notification triggers: All
    """

    type: Literal["telegram"] = "telegram"
    """
    Type value associated with this kind of connection.
    """

    bot_token: Password
    """
    The bot token assigned to the Sonarr instance.
    """

    chat_id: NonEmptyStr
    """
    The ID of the chat room to send messages to.

    You must start a conversation with the bot or add it to your group to receive messages.
    """

    send_silently: bool = False
    """
    Sends the message silently. Users will receive a notification with no sound.
    """

    _implementation: str = "Telegram"
    _implementation_name: str = "Telegram"
    _config_contract: str = "TelegramSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("bot_token", "botToken", {"is_field": True}),
        ("chat_id", "chatId", {"is_field": True}),
        ("send_silently", "sendSilently", {"is_field": True}),
    ]


class TraktConnection(Connection):
    """
    Send media update and health alert messages to the Trakt media tracker.

    !!! note

        Sonarr directly authenticates with Trakt to generate tokens for it to use.
        At the moment, the easiest way to generate the tokens for Buildarr
        is to do it using the GUI within Sonarr, and use the following
        shell command to retrieve the generated configuration.

        ```bash
        $ curl -X "GET" "<sonarr-url>/api/v3/notification" -H "X-Api-Key: <api-key>"
        ```

    Supported notification triggers: All except `on_grab`, `on_rename`,
    `on_health_issue` and `on_application_update`
    """

    # FIXME: Determine easier procedure for getting access tokens and test.

    type: Literal["trakt"] = "trakt"
    """
    Type value associated with this kind of connection.
    """

    access_token: Password
    """
    Access token for Sonarr from Trakt.
    """

    refresh_token: Password
    """
    Refresh token for Sonarr from Trakt.
    """

    expires: datetime
    """
    Expiry date-time of the access token, preferably in ISO-8601 format and in UTC.

    Example: `2023-05-10T15:34:08.117451Z`
    """

    auth_user: TraktAuthUser
    """
    The username being authenticated in Trakt.
    """

    _implementation_name: str = "Trakt"
    _implementation: str = "Trakt"
    _config_contract: str = "TraktSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("access_token", "accessToken", {"is_field": True}),
        ("refresh_token", "refreshToken", {"is_field": True}),
        ("expires", "expires", {"is_field": True, "encoder": trakt_expires_encoder}),
        ("auth_user", "authUser", {"is_field": True}),
    ]


class TwitterConnection(Connection):
    """
    Send media update and health alert messages via Twitter.

    Twitter requires you to create an application for their API
    to generate the necessary keys and secrets.
    If unsure how to proceed, refer to these guides from
    [Twitter](https://developer.twitter.com/en/docs/authentication/oauth-1-0a/api-key-and-secret)
    and [WikiArr](https://wiki.servarr.com/useful-tools#twitter-connect).

    Access tokens can be obtained using the prodecure documented [here](
    https://developer.twitter.com/en/docs/authentication/oauth-1-0a/obtaining-user-access-tokens).

    Supported notification triggers: All except `on_rename`
    """

    type: Literal["twitter"] = "twitter"
    """
    Type value associated with this kind of connection.
    """

    consumer_key: Password
    """
    Consumer key from a Twitter application.
    """

    consumer_secret: Password
    """
    Consumer key from a Twitter application.
    """

    access_token: Password
    """
    Access token for a Twitter user.
    """

    access_token_secret: Password
    """
    Access token secret for a Twitter user.
    """

    mention: NonEmptyStr
    """
    Mention this user in sent tweets.
    """

    direct_message: bool = True
    """
    Send a direct message instead of a public message.
    """

    _implementation_name: str = "Twitter"
    _implementation: str = "Twitter"
    _config_contract: str = "TwitterSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("consumer_key", "consumerKey", {"is_field": True}),
        ("consumer_secret", "consumerSecret", {"is_field": True}),
        ("access_token", "accessToken", {"is_field": True}),
        ("access_token_secret", "accessTokenSecret", {"is_field": True}),
        ("mention", "mention", {"is_field": True}),
        ("direct_message", "direct_message", {"is_field": True}),
    ]


class WebhookConnection(Connection):
    """
    Send media update and health alert notifications to a webhook API.

    Supported notification triggers: All
    """

    type: Literal["webhook"] = "webhook"
    """
    Type value associated with this kind of connection.
    """

    url: AnyHttpUrl
    """
    Webhook URL to send notifications to.
    """

    method: WebhookMethod = WebhookMethod.POST
    """
    HTTP request method type to use.

    Values:

    * `POST`
    * `PUT`
    """

    username: NonEmptyStr
    """
    Webhook API username.
    """

    password: Password
    """
    Webhook API password.
    """

    _implementation_name: str = "Webhook"
    _implementation: str = "Webhook"
    _config_contract: str = "WebhookSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("url", "url", {"is_field": True}),
        ("method", "method", {"is_field": True}),
        ("username", "username", {"is_field": True}),
        ("password", "password", {"is_field": True}),
    ]


CONNECTION_TYPES: Tuple[Type[Connection], ...] = (
    BoxcarConnection,
    CustomscriptConnection,
    DiscordConnection,
    EmailConnection,
    EmbyConnection,
    GotifyConnection,
    JoinConnection,
    KodiConnection,
    MailgunConnection,
    PlexHomeTheaterConnection,
    PlexMediaCenterConnection,
    PlexMediaServerConnection,
    ProwlConnection,
    PushbulletConnection,
    PushoverConnection,
    SendgridConnection,
    SlackConnection,
    SynologyIndexerConnection,
    TelegramConnection,
    TraktConnection,
    TwitterConnection,
    WebhookConnection,
)

CONNECTION_TYPE_MAP: Dict[str, Type[Connection]] = {
    connection_type._implementation: connection_type for connection_type in CONNECTION_TYPES
}

ConnectionType = Union[
    BoxcarConnection,
    CustomscriptConnection,
    DiscordConnection,
    EmailConnection,
    EmbyConnection,
    GotifyConnection,
    JoinConnection,
    KodiConnection,
    MailgunConnection,
    PlexHomeTheaterConnection,
    PlexMediaCenterConnection,
    PlexMediaServerConnection,
    ProwlConnection,
    PushbulletConnection,
    PushoverConnection,
    SendgridConnection,
    SlackConnection,
    SynologyIndexerConnection,
    TelegramConnection,
    TraktConnection,
    TwitterConnection,
    WebhookConnection,
]


class SonarrConnectSettingsConfig(SonarrConfigBase):
    """
    Manage notification connections in Sonarr.
    """

    delete_unmanaged: bool = False
    """
    Automatically delete connections not configured in Buildarr.

    Take care when enabling this option, as this can remove connections automatically
    managed by other applications.
    """

    definitions: Dict[str, Annotated[ConnectionType, Field(discriminator="type")]] = {}
    """
    Connection definitions to configure in Sonarr.
    """

    @classmethod
    def from_remote(cls, secrets: SonarrSecrets) -> Self:
        connections = api_get(secrets, "/api/v3/notification")
        tag_ids: Dict[str, int] = (
            {tag["label"]: tag["id"] for tag in api_get(secrets, "/api/v3/tag")}
            if any(connection["tags"] for connection in connections)
            else {}
        )
        return cls(
            definitions={
                connection["name"]: CONNECTION_TYPE_MAP[connection["implementation"]]._from_remote(
                    tag_ids=tag_ids,
                    remote_attrs=connection,
                )
                for connection in connections
            },
        )

    def update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> bool:
        #
        changed = False
        #
        connection_ids: Dict[str, int] = {
            connection_json["name"]: connection_json["id"]
            for connection_json in api_get(secrets, "/api/v3/notification")
        }
        tag_ids: Dict[str, int] = (
            {tag["label"]: tag["id"] for tag in api_get(secrets, "/api/v3/tag")}
            if any(connection.tags for connection in self.definitions.values())
            or any(connection.tags for connection in remote.definitions.values())
            else {}
        )
        #
        for connection_name, connection in self.definitions.items():
            connection_tree = f"{tree}.definitions[{repr(connection_name)}]"
            #
            if connection_name not in remote.definitions:
                connection._create_remote(
                    tree=connection_tree,
                    secrets=secrets,
                    tag_ids=tag_ids,
                    connection_name=connection_name,
                )
                changed = True
            #
            elif connection._update_remote(
                tree=connection_tree,
                secrets=secrets,
                remote=remote.definitions[connection_name],  # type: ignore[arg-type]
                tag_ids=tag_ids,
                connection_id=connection_ids[connection_name],
                connection_name=connection_name,
            ):
                changed = True
        #
        for connection_name, connection in remote.definitions.items():
            if connection_name not in self.definitions:
                connection_tree = f"{tree}.definitions[{repr(connection_name)}]"
                if self.delete_unmanaged:
                    connection._delete_remote(
                        tree=connection_tree,
                        secrets=secrets,
                        connection_id=connection_ids[connection_name],
                    )
                    changed = True
                else:
                    logger.debug("%s: (...) (unmanaged)", connection_tree)
        #
        return changed



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin release profile configuration.
"""


from __future__ import annotations

import json

from logging import getLogger
from pathlib import Path
from typing import Any, Dict, Iterable, List, Mapping, Optional, Set, cast

from buildarr.config import ConfigTrashIDNotFoundError, RemoteMapEntry
from buildarr.types import NonEmptyStr, TrashID
from pydantic import validator
from typing_extensions import Self

from ...api import api_delete, api_get, api_post, api_put
from ...secrets import SonarrSecrets
from ..types import SonarrConfigBase

logger = getLogger(__name__)


class TrashFilter(SonarrConfigBase):
    """
    Defines various ways that release profile terms from the guide are synchronised with Sonarr.
    These terms have individual trash IDs, and using this filter allows you to
    pick and choose which parts of the release profile you want to use.

    ```yaml
    ...
      release_profiles:
        definitions:
          Optionals:
            trash_id: 76e060895c5b8a765c310933da0a5357
            filter:
              include:
                - ea83f4740cec4df8112f3d6dd7c82751 # Prefer Season Packs
                - cec8880b847dd5d31d29167ee0112b57 # Ignore 720p/1080p HEVC (Golden Rule)
                # All the other preferred words in the 'Optionals' profile
                # are ignored.
    ```

    The attributes `include` and `exclude` are mutually exclusive.
    If both are defined, `include` is used.
    """

    # TODO: mutual exclusion

    include: List[TrashID] = []
    """
    A list of `trash_id` values representing terms (`Required`, `Ignored`, or `Preferred`)
    that should be included in the created Release Profile in Sonarr.

    Terms that are *not* specified here are excluded automatically.
    """

    exclude: List[TrashID] = []
    """
    A list of `trash_id` values representing terms (`Required`, `Ignored`, or `Preferred`)
    that should be excluded from the created Release Profile in Sonarr.

    Terms that are *not* specified here are included automatically.
    """


class PreferredWord(SonarrConfigBase):
    """
    Buildarr object representing a Preferred Word in a Release Profile.
    """

    term: NonEmptyStr
    score: int


class ReleaseProfile(SonarrConfigBase):
    # Release profile data structure.
    #
    # There are two types of release profile: the type where filters are manually defined,
    # and the type where the profile is simply a reference to a pre-made
    # TRaSH-Guides profile (with some handling options).

    enable: bool = True
    """
    Enable the release profile in Sonarr.

    If set to `False`, the release profile will be uploaded to Sonarr, but inactive.
    """

    trash_id: Optional[TrashID] = None
    """
    The trash ID of the release profile to import from TRaSH-Guides metadata.

    If not specified, Buildarr assumes the release profile filters are manually defined.
    """

    # Preferred word Trash ID filter for this release profile.
    # See the `TrashFilter` class for more details.
    # If undefined, use the TRaSH-Guides provided release profile defaults.
    filter: TrashFilter = TrashFilter()
    """
    ###### ::: buildarr_sonarr.config.profiles.release.TrashFilter
        options:
          members:
            - include
            - exclude
          show_root_heading: false
          show_source: false
    """

    strict_negative_scores: bool = False
    """
    Enables preferred term scores less than 0 to be instead treated
    as "Must Not Contain" (ignored) terms.

    For example, if something is "Preferred" with a score of `-10,` it will instead
    be put in the "Must Not Contain" section of the uploaded release profile.
    """

    # TODO: constraint that checks at least one of these filters
    #       are not empty, if trash_id is not defined.

    must_contain: Set[NonEmptyStr] = set()
    """
    A list of terms to mark as "Must Contain" in Sonarr.

    The release will be rejected if it does not contain one or more of terms
    (case insensitive) or regular expressions.
    """

    must_not_contain: Set[NonEmptyStr] = set()
    """
    A list of terms to mark as "Must Not Contain" in Sonarr.

    The release will be rejected if it contains one or more of terms (case insensitive)
    or regular expressions.
    """

    preferred: List[PreferredWord] = []
    """
    Assign a score to terms (or a regular expression patterns) found in the file name
    of a release, and prefer it or not prefer it depending on the result
    of comparing a release file name against all terms.

    ```yaml
    ...
      release_profiles:
        Example:
          preferred:
            - term: "/\b(amzn|amazon)\b(?=[ ._-]web[ ._-]?(dl|rip)\b)/i"
              score: 100
            - term: "/(-BRiNK|-CHX|-GHOSTS|-EVO|)\b/i"
              score: -10000
    ```

    The release will be preferred based on each term's score.
    Positive scores will be more preferred, and negative scores will be less preferred.
    """

    include_preferred_when_renaming: bool = False
    """
    Add preferred words to the file name as `{Preferred Words}`
    when doing automatic renaming in Sonarr.
    """

    # None = (Any)
    indexer: Optional[NonEmptyStr] = None
    """
    When set to `None` or an empty string, use any available indexer.
    """

    #
    tags: Set[NonEmptyStr] = set()
    """
    A list of one or more strings representing tags that will be applied to this release profile.

    All tags on an existing release profile (if present) are removed
    and replaced with only the tags in this list.

    If an empty list is explicitly defined, no tags will be set on the release profile,
    and any existing tags (if present) are removed.
    """

    @validator("preferred")
    def sort_preferred(cls, preferred: Iterable[PreferredWord]) -> List[PreferredWord]:
        return sorted(preferred, key=lambda p: (-p.score, p.term))

    @classmethod
    def _get_remote_map(
        cls,
        indexer_ids: Mapping[str, int],
        tag_ids: Mapping[str, int],
    ) -> List[RemoteMapEntry]:
        return [
            ("enable", "enabled", {}),
            ("must_contain", "required", {"encoder": lambda v: sorted(v)}),
            ("must_not_contain", "ignored", {"encoder": lambda v: sorted(v)}),
            (
                "preferred",
                "preferred",
                {
                    "decoder": lambda v: [
                        PreferredWord(term=p["key"], score=p["value"]) for p in v
                    ],
                    "encoder": lambda v: [{"key": p.term, "value": p.score} for p in v],
                },
            ),
            ("include_preferred_when_renaming", "includePreferredWhenRenaming", {}),
            (
                "indexer",
                "indexerId",
                {
                    "decoder": lambda v: next(
                        (ind for ind, ind_id in indexer_ids.items() if ind_id == v),
                        None,
                    ),
                    "encoder": lambda v: indexer_ids[v] if v else 0,
                },
            ),
            (
                "tags",
                "tags",
                {
                    "decoder": lambda v: set(
                        (tag for tag, tag_id in tag_ids.items() if tag_id in v),
                    ),
                    "encoder": lambda v: sorted(tag_ids[tag] for tag in v),
                },
            ),
        ]

    @classmethod
    def _from_remote(
        cls,
        indexer_ids: Mapping[str, int],
        tag_ids: Mapping[str, int],
        remote_attrs: Mapping[str, Any],
    ) -> Self:
        return cls(**cls.get_local_attrs(cls._get_remote_map(indexer_ids, tag_ids), remote_attrs))

    @property
    def uses_trash_metadata(self) -> bool:
        """
        A flag determining whether or not this configuration uses TRaSH-Guides metadata.

        Returns:
            `True` if TRaSH-Guides metadata is used, otherwise `False`
        """
        return bool(self.trash_id)

    def _render_trash_metadata(self, trash_metadata_dir: Path) -> None:
        """
        Render configuration attributes obtained from TRaSH-Guides, in-place.

        Args:
            trash_metadata_dir (Path): TRaSH-Guides metadata directory.
        """
        if not self.trash_id:
            return
        for profile_file in (trash_metadata_dir / "docs" / "json" / "sonarr" / "rp").iterdir():
            with profile_file.open() as f:
                profile: Dict[str, Any] = json.load(f)
                if cast(str, profile["trash_id"]).lower() == self.trash_id:
                    #
                    must_contain: Set[str] = set()
                    must_not_contain: Set[str] = set()
                    preferred: List[PreferredWord] = []
                    #
                    for required in profile["required"]:
                        if isinstance(required, dict):
                            term: str = required["term"]
                            if (
                                "trash_id" in required
                                and cast(str, required["trash_id"]).lower() in self.filter.exclude
                            ):
                                must_not_contain.add(term)
                            else:
                                must_contain.add(term)
                        else:
                            must_contain.add(required)
                    #
                    for ignored in profile["ignored"]:
                        if isinstance(ignored, dict):
                            term = ignored["term"]
                            if (
                                "trash_id" in ignored
                                and cast(str, ignored["trash_id"]).lower() in self.filter.include
                            ):
                                must_contain.add(term)
                            else:
                                must_not_contain.add(term)
                        else:
                            must_not_contain.add(ignored)
                    #
                    for pref in profile["preferred"]:
                        score: int = pref["score"]
                        for term_obj in pref["terms"]:
                            term_trash_id: Optional[str] = None
                            if isinstance(term_obj, dict):
                                term = term_obj["term"]
                                if "trash_id" in term_obj:
                                    term_trash_id = cast(str, term_obj["trash_id"]).lower()
                            else:
                                term = term_obj
                            if (self.strict_negative_scores and score < 0) or (
                                term_trash_id and term_trash_id in self.filter.exclude
                            ):
                                must_not_contain.add(term)
                            else:
                                preferred.append(PreferredWord(term=term, score=score))
                    #
                    self.must_contain = must_contain  # type: ignore[assignment]
                    self.must_not_contain = must_not_contain  # type: ignore[assignment]
                    self.preferred = preferred
                    return
        raise ConfigTrashIDNotFoundError(
            f"Unable to find Sonarr release profile file with trash ID '{self.trash_id}'",
        )

    def _create_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        profile_name: str,
        indexer_ids: Mapping[str, int],
        tag_ids: Mapping[str, int],
    ) -> None:
        api_post(
            secrets,
            "/api/v3/releaseprofile",
            {
                "name": profile_name,
                **self.get_create_remote_attrs(tree, self._get_remote_map(indexer_ids, tag_ids)),
            },
        )

    def _update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        profile_id: int,
        profile_name: str,
        indexer_ids: Mapping[str, int],
        tag_ids: Mapping[str, int],
    ) -> bool:
        changed, remote_attrs = self.get_update_remote_attrs(
            tree,
            remote,
            self._get_remote_map(indexer_ids, tag_ids),
            check_unmanaged=True,
            set_unchanged=True,
        )
        if changed:
            api_put(
                secrets,
                f"/api/v3/releaseprofile/{profile_id}",
                {"id": profile_id, "name": profile_name, **remote_attrs},
            )
            return True
        return False

    def _delete_remote(self, tree: str, secrets: SonarrSecrets, profile_id: int) -> None:
        logger.info("%s: (...) -> (deleted)", tree)
        api_delete(secrets, f"/api/v3/releaseprofile/{profile_id}")

    # Tell Pydantic to validate in-place assignments of attributes.
    # This ensures that any validators that parse attributes to consistent values run.
    class Config(SonarrConfigBase.Config):
        validate_assignment = True


class SonarrReleaseProfilesSettingsConfig(SonarrConfigBase):
    """
    Configuration parameters for controlling how Buildarr handles release profiles.
    """

    delete_unmanaged: bool = False
    """
    Automatically delete release profiles not defined in Buildarr.
    """

    definitions: Dict[str, ReleaseProfile] = {}
    """
    Define release profiles to configure on Sonarr here.

    If there are no release profiles defined and `delete_unmanaged` is `False`,
    Buildarr will not modify existing release profiles, but if `delete_unmanaged` is `True`,
    **Buildarr will delete all existing profiles. Be careful when using `delete_unmanaged`.**
    """

    @classmethod
    def from_remote(cls, secrets: SonarrSecrets) -> Self:
        profiles: List[Dict[str, Any]] = api_get(secrets, "/api/v3/releaseprofile")
        indexer_ids: Dict[str, int] = (
            {tag["name"]: tag["id"] for tag in api_get(secrets, "/api/v3/indexer")}
            if any(profile["indexerId"] for profile in profiles)
            else {}
        )
        tag_ids: Dict[str, int] = (
            {tag["label"]: tag["id"] for tag in api_get(secrets, "/api/v3/tag")}
            if any(profile["tags"] for profile in profiles)
            else {}
        )
        return cls(
            definitions={
                profile["name"]: ReleaseProfile._from_remote(indexer_ids, tag_ids, profile)
                for profile in profiles
            },
        )

    def update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> bool:
        #
        changed = False
        #
        profile_ids: Dict[str, int] = {
            profile_json["name"]: profile_json["id"]
            for profile_json in api_get(secrets, "/api/v3/releaseprofile")
        }
        indexer_ids: Dict[str, int] = (
            {tag["name"]: tag["id"] for tag in api_get(secrets, "/api/v3/indexer")}
            if any(p.indexer for p in self.definitions.values())
            or any(p.indexer for p in remote.definitions.values())
            else {}
        )
        tag_ids: Dict[str, int] = (
            {tag["label"]: tag["id"] for tag in api_get(secrets, "/api/v3/tag")}
            if any(profile.tags for profile in self.definitions.values())
            or any(profile.tags for profile in remote.definitions.values())
            else {}
        )
        #
        for profile_name, profile in self.definitions.items():
            profile_tree = f"{tree}.definitions[{repr(profile_name)}]"
            #
            if profile_name not in remote.definitions:
                profile._create_remote(
                    tree=profile_tree,
                    secrets=secrets,
                    profile_name=profile_name,
                    indexer_ids=indexer_ids,
                    tag_ids=tag_ids,
                )
                changed = True
            #
            elif profile._update_remote(
                tree=profile_tree,
                secrets=secrets,
                remote=remote.definitions[profile_name],
                profile_id=profile_ids[profile_name],
                profile_name=profile_name,
                indexer_ids=indexer_ids,
                tag_ids=tag_ids,
            ):
                changed = True
        #
        for profile_name, profile in remote.definitions.items():
            if profile_name not in self.definitions:
                profile_tree = f"{tree}.definitions[{repr(profile_name)}]"
                if self.delete_unmanaged:
                    profile._delete_remote(
                        tree=profile_tree,
                        secrets=secrets,
                        profile_id=profile_ids[profile_name],
                    )
                    changed = True
                else:
                    logger.debug("%s: (...) (unmanaged)", profile_tree)
        #
        return changed



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin download client definition.
"""


from __future__ import annotations

from logging import getLogger
from typing import Any, Dict, List, Literal, Mapping, Optional, Set, Tuple, Type

from buildarr.config import RemoteMapEntry
from buildarr.types import BaseEnum, NonEmptyStr, Password, Port
from typing_extensions import Self

from ...api import api_delete, api_post, api_put
from ...secrets import SonarrSecrets
from ..types import SonarrConfigBase

logger = getLogger(__name__)


class NzbgetPriority(BaseEnum):
    """
    NZBGet media priority.

    Values:

    * `verylow` (Very Low)
    * `low` (Low)
    * `normal` (Normal)
    * `high` (High)
    * `veryhigh` (Very High)
    * `force` (Force)
    """

    verylow = -100
    low = -50
    normal = 0
    high = 50
    veryhigh = 100
    force = 900


class NzbvortexPriority(BaseEnum):
    """
    NZBVortex media priority.

    Values:

    * `low` (Low)
    * `normal` (Normal)
    * `high` (High)
    """

    low = -1
    normal = 0
    high = 1


class SabnzbdPriority(BaseEnum):
    """
    SABnzbd media priority.

    Values:

    * `default` (Default)
    * `paused` (Paused)
    * `low` (Low)
    * `normal` (Normal)
    * `high` (High)
    * `force` (Force)
    """

    default = -100
    paused = -2
    low = -1
    normal = 0
    high = 1
    force = 2


class DelugePriority(BaseEnum):
    """
    Deluge queue priority.

    Values:

    * `last` (Last)
    * `first` (First)
    """

    last = 0
    first = 1


class FloodMediaTag(BaseEnum):
    """
    Type of tag to set on media within Flood.

    Multiple can be specified at a time.

    Values:

    * `title-slug` (Title Slug)
    * `quality` (Quality)
    * `language` (Language)
    * `release-group` (Release Group)
    * `year` (Year)
    * `indexer` (Indexer)
    * `network` (Network)
    """

    title_slug = 0
    quality = 1
    language = 2
    release_group = 3
    year = 4
    indexer = 5
    network = 6


class QbittorrentPriority(BaseEnum):
    """
    qBittorrent queue priority.

    Values:

    * `last` (Last)
    * `first` (First)
    """

    last = 0
    first = 1


class QbittorrentState(BaseEnum):
    """
    qBittorrent initial state.

    Values:

    * `start` (Start)
    * `force-start` (Force Start)
    * `pause` (Pause)
    """

    start = 0
    force_start = 1
    pause = 2


class RtorrentPriority(BaseEnum):
    """
    RTorrent media priority.

    Values:

    * `verylow` (Very Low)
    * `low` (Low)
    * `normal` (Normal)
    * `high` (High)
    """

    verylow = 0
    low = 1
    normal = 2
    high = 3


class TransmissionPriority(BaseEnum):
    """
    Transmission queue priority.

    Values:

    * `last` (Last)
    * `first` (First)
    """

    last = 0
    first = 1


class UtorrentPriority(BaseEnum):
    """
    uTorrent queue priority.

    Values:

    * `last` (Last)
    * `first` (First)
    """

    last = 0
    first = 1


class UtorrentState(BaseEnum):
    """
    uTorrent initial state.

    Values:

    * `start` (Start)
    * `force-start` (Force Start)
    * `pause` (Pause)
    * `stop` (Stop)
    """

    start = 0
    force_start = 1
    pause = 2
    stop = 3


class DownloadClient(SonarrConfigBase):
    """
    Download clients are defined using the following format.
    Here is an example of a Transmission download client being configured.

    ```yaml
    ---

    sonarr:
      settings:
        download_clients:
          definitions:
            Transmission: # Name of the download client
              type: "transmission" # Type of download client
              enable: true # Enable the download client in Sonarr
              host: "transmission"
              port: 9091
              category: "sonarr"
              # Define any other type-specific or global
              # download client attributes as needed.
    ```

    Every download client definition must have the correct `type` value defined,
    to tell Buildarr what type of download client to configure.
    The name of the download client definition is just a name, and has no meaning.

    `enable` can be set to `False` to keep the download client configured on Sonarr,
    but disabled so that it is inactive.

    The below attributes can be defined on any type of download client.
    """

    enable: bool = True
    """
    When `True`, this download client is active and Sonarr is able to send requests to it.
    """

    priority: int = 1
    """
    Download client priority.

    Clients with a lower value are prioritised.
    Round-robin is used for clients with the same priority.
    """

    remove_completed_downloads: bool = True
    """
    Remove completed downloads from the download client history.

    For torrents, this happens once seeding is complete.
    """

    remove_failed_downloads: bool = True
    """
    Remove failed downloads from the download client history. (Usenet clients only)
    """

    tags: Set[NonEmptyStr] = set()
    """
    Sonarr tags to assign to the download clients.
    Only media under those tags will be assigned to this client.

    If no tags are assigned, all media can use the client.
    """

    _implementation_name: str
    _implementation: str
    _config_contract: str
    _remote_map: List[RemoteMapEntry]

    @classmethod
    def _get_base_remote_map(cls, tag_ids: Mapping[str, int]) -> List[RemoteMapEntry]:
        return [
            ("enable", "enable", {}),
            ("priority", "priority", {}),
            ("remove_completed_downloads", "removeCompletedDownloads", {}),
            ("remove_failed_downloads", "removeFailedDownloads", {}),
            (
                "tags",
                "tags",
                {
                    "decoder": lambda v: set(
                        (tag for tag, tag_id in tag_ids.items() if tag_id in v),
                    ),
                    "encoder": lambda v: sorted(tag_ids[tag] for tag in v),
                },
            ),
        ]

    @classmethod
    def _from_remote(cls, tag_ids: Mapping[str, int], remote_attrs: Mapping[str, Any]) -> Self:
        return cls(
            **cls.get_local_attrs(
                cls._get_base_remote_map(tag_ids) + cls._remote_map,
                remote_attrs,
            ),
        )

    def _create_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        tag_ids: Mapping[str, int],
        downloadclient_name: str,
    ) -> None:
        api_post(
            secrets,
            "/api/v3/downloadclient",
            {
                "name": downloadclient_name,
                "implementation": self._implementation,
                "implementationName": self._implementation_name,
                "configContract": self._config_contract,
                **self.get_create_remote_attrs(
                    tree,
                    self._get_base_remote_map(tag_ids) + self._remote_map,
                ),
            },
        )

    def _update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        tag_ids: Mapping[str, int],
        downloadclient_id: int,
        downloadclient_name: str,
    ) -> bool:
        updated, remote_attrs = self.get_update_remote_attrs(
            tree,
            remote,
            self._get_base_remote_map(tag_ids) + self._remote_map,
            set_unchanged=True,
            # TODO: Should we also enable check_unmanaged here?
        )
        if updated:
            api_put(
                secrets,
                f"/api/v3/downloadclient/{downloadclient_id}",
                {
                    "id": downloadclient_id,
                    "name": downloadclient_name,
                    "implementation": self._implementation,
                    "implementationName": self._implementation_name,
                    "configContract": self._config_contract,
                    **remote_attrs,
                },
            )
            return True
        return False

    def _delete_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        downloadclient_id: int,
    ) -> None:
        logger.info("%s: (...) -> (deleted)", tree)
        api_delete(secrets, f"/api/v3/downloadclient/{downloadclient_id}")


class UsenetDownloadClient(DownloadClient):
    """
    Usenet-based download client.
    """

    pass


class TorrentDownloadClient(DownloadClient):
    """
    Torrent-based download client.
    """

    pass


class DownloadstationUsenetDownloadClient(UsenetDownloadClient):
    """
    Download client which uses Usenet via Download Station.
    """

    type: Literal["downloadstation-usenet"] = "downloadstation-usenet"
    """
    Type value associated with this kind of download client.
    """

    host: NonEmptyStr
    """
    Download Station host name.
    """

    port: Port = 5000  # type: ignore[assignment]
    """
    Download client access port.
    """

    use_ssl: bool = False
    """
    Use a secure connection when connecting to the download client.
    """

    username: NonEmptyStr
    """
    User name to use when authenticating with the download client.
    """

    password: Password
    """
    Password to use to authenticate the download client user.
    """

    category: Optional[str] = None
    """
    Associate media from Sonarr with a category.
    Creates a `[category]` subdirectory in the output directory.

    Adding a category specific to Sonarr avoids conflicts with unrelated non-Sonarr downloads.
    Using a category is optional, but strongly recommended.
    """

    directory: Optional[str] = None
    """
    Optional shared folder to put downloads into.

    Leave blank, set to `null` or undefined to use the default download client location.
    """

    _implementation_name: str = "Download Station"
    _implementation: str = "UsenetDownloadStation"
    _config_contract: str = "DownloadStationSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        ("username", "username", {"is_field": True}),
        ("password", "password", {"is_field": True}),
        (
            "category",
            "tvCategory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        (
            "category",
            "tvDirectory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
    ]


class NzbgetDownloadClient(UsenetDownloadClient):
    """
    NZBGet download client.
    """

    type: Literal["nzbget"] = "nzbget"
    """
    Type value associated with this kind of download client.
    """

    host: NonEmptyStr
    """
    NZBGet host name.
    """

    port: Port = 5000  # type: ignore[assignment]
    """
    Download client access port.
    """

    use_ssl: bool = False
    """
    Use a secure connection when connecting to the download client.
    """

    url_base: Optional[str] = None
    """
    Adds a prefix to the NZBGet url, e.g. `http://[host]:[port]/[url_base]/jsonrpc`.
    """

    username: NonEmptyStr
    """
    User name to use when authenticating with the download client.
    """

    password: Password
    """
    Password to use to authenticate the download client user.
    """

    category: Optional[str] = None
    """
    Associate media from Sonarr with a category.

    Adding a category specific to Sonarr avoids conflicts with unrelated non-Sonarr downloads.
    Using a category is optional, but strongly recommended.
    """

    recent_priority: NzbgetPriority = NzbgetPriority.normal
    """
    Priority to use when grabbing episodes that aired within the last 14 days.

    Values:

    * `verylow`
    * `low`
    * `normal`
    * `high`
    * `veryhigh`
    * `force`
    """

    older_priority: NzbgetPriority = NzbgetPriority.normal
    """
    Priority to use when grabbing episodes that aired over 14 days ago.

    Values:

    * `verylow`
    * `low`
    * `normal`
    * `high`
    * `veryhigh`
    * `force`
    """

    add_paused: bool = False
    """
    Add media to the download client in the paused state.

    This option requires NZBGet version 16.0 or later.
    """

    _implementation_name: str = "NZBGet"
    _implementation: str = "Nzbget"
    _config_contract: str = "NzbgetSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        (
            "url_base",
            "urlBase",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("username", "username", {"is_field": True}),
        ("password", "password", {"is_field": True}),
        (
            "category",
            "tvCategory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("recent_priority", "recentTvPriority", {"is_field": True}),
        ("older_priority", "olderTvPriority", {"is_field": True}),
        ("add_paused", "addPaused", {"is_field": True}),
    ]


class NzbvortexDownloadClient(UsenetDownloadClient):
    """
    NZBVortex download client.
    """

    type: Literal["nzbvortex"] = "nzbvortex"
    """
    Type value associated with this kind of download client.
    """

    host: NonEmptyStr
    """
    NZBVortex host name.
    """

    port: Port = 4321  # type: ignore[assignment]
    """
    Download client access port.
    """

    use_ssl: bool = False
    """
    Use a secure connection when connecting to the download client.
    """

    url_base: Optional[str] = None
    """
    Adds a prefix to the NZBVortex url, e.g. `http://[host]:[port]/[url_base]/api`.
    """

    api_key: Password
    """
    API key to use to authenticate with the download client.
    """

    category: Optional[str] = None
    """
    Associate media from Sonarr with a category.

    Adding a category specific to Sonarr avoids conflicts with unrelated non-Sonarr downloads.
    Using a category is optional, but strongly recommended.
    """

    recent_priority: NzbvortexPriority = NzbvortexPriority.normal
    """
    Priority to use when grabbing episodes that aired within the last 14 days.

    Values:

    * `low`
    * `normal`
    * `high`
    """

    older_priority: NzbvortexPriority = NzbvortexPriority.normal
    """
    Priority to use when grabbing episodes that aired over 14 days ago.

    Values:

    * `low`
    * `normal`
    * `high`
    """

    _implementation_name: str = "NZBVortex"
    _implementation: str = "NzbVortex"
    _config_contract: str = "NzbVortexSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        (
            "url_base",
            "urlBase",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("api_key", "apiKey", {"is_field": True}),
        (
            "category",
            "tvCategory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("recent_priority", "recentTvPriority", {"is_field": True}),
        ("older_priority", "olderTvPriority", {"is_field": True}),
    ]


class PneumaticDownloadClient(UsenetDownloadClient):
    """
    Download client for the Pneumatic NZB add-on for Kodi/XMBC.
    """

    type: Literal["pneumatic"] = "pneumatic"
    """
    Type value associated with this kind of download client.
    """

    nzb_folder: NonEmptyStr
    """
    Folder in which Sonarr will store `.nzb` files.

    This folder will need to be reachable from Kodi/XMBC.
    """

    strm_folder: NonEmptyStr
    """
    Folder from which `.strm` files will be imported by Drone.
    """

    _implementation_name: str = "Pneumatic"
    _implementation: str = "Pneumatic"
    _config_contract: str = "PneumaticSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("nzb_folder", "nzbFolder", {"is_field": True}),
        ("strm_folder", "strmFolder", {"is_field": True}),
    ]


class SabnzbdDownloadClient(UsenetDownloadClient):
    """
    SABnzbd download client.
    """

    type: Literal["sabnzbd"] = "sabnzbd"
    """
    Type value associated with this kind of download client.
    """

    host: NonEmptyStr
    """
    SABnzbd host name.
    """

    port: Port = 4321  # type: ignore[assignment]
    """
    Download client access port.
    """

    use_ssl: bool = False
    """
    Use a secure connection when connecting to the download client.
    """

    url_base: Optional[str] = None
    """
    Adds a prefix to the SABnzbd URL, e.g. `http://[host]:[port]/[url_base]/api/`.
    """

    api_key: Password
    """
    API key to use to authenticate with the download client.
    """

    category: Optional[str] = None
    """
    Associate media from Sonarr with a category.

    Adding a category specific to Sonarr avoids conflicts with unrelated non-Sonarr downloads.
    Using a category is optional, but strongly recommended.
    """

    recent_priority: SabnzbdPriority = SabnzbdPriority.default
    """
    Priority to use when grabbing episodes that aired within the last 14 days.

    Values:

    * `default`
    * `paused`
    * `low`
    * `normal`
    * `high`
    * `force`
    """

    older_priority: SabnzbdPriority = SabnzbdPriority.default
    """
    Priority to use when grabbing episodes that aired over 14 days ago.

    Values:

    * `default`
    * `paused`
    * `low`
    * `normal`
    * `high`
    * `force`
    """

    _implementation_name: str = "SABnzbd"
    _implementation: str = "Sabnzbd"
    _config_contract: str = "SabnzbdSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        (
            "url_base",
            "urlBase",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("api_key", "apiKey", {"is_field": True}),
        (
            "category",
            "tvCategory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("recent_priority", "recentTvPriority", {"is_field": True}),
        ("older_priority", "olderTvPriority", {"is_field": True}),
    ]


class UsenetBlackholeDownloadClient(UsenetDownloadClient):
    """
    Usenet Blackhole download client.
    """

    type: Literal["usenet-blackhole"] = "usenet-blackhole"
    """
    Type value associated with this kind of download client.
    """

    nzb_folder: NonEmptyStr
    """
    Folder in which Sonarr will store `.nzb` files.
    """

    watch_folder: NonEmptyStr
    """
    Folder from which Sonarr should import completed downloads.
    """

    _implementation_name: str = "Usenet Blackhole"
    _implementation: str = "UsenetBlackhole"
    _config_contract: str = "UsenetBlackholeSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("nzb_folder", "nzbFolder", {"is_field": True}),
        ("watch_folder", "watchFolder", {"is_field": True}),
    ]


class Aria2DownloadClient(TorrentDownloadClient):
    """
    Aria2 download client.
    """

    type: Literal["aria2"] = "aria2"
    """
    Type value associated with this kind of download client.
    """

    host: NonEmptyStr
    """
    Aria2 host name.
    """

    port: Port = 6800  # type: ignore[assignment]
    """
    Download client access port.
    """

    use_ssl: bool = False
    """
    Use a secure connection when connecting to the download client.
    """

    rpc_path: NonEmptyStr = "/rpc"  # type: ignore[assignment]
    """
    XML RPC path in the Aria2 client URL.
    """

    secret_token: Password
    """
    Secret token to use to authenticate with the download client.
    """

    _implementation_name: str = "Aria2"
    _implementation: str = "Aria2"
    _config_contract: str = "Aria2Settings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        ("rpc_path", "rpcPath", {"is_field": True}),
        ("secret_token", "secretToken", {"is_field": True}),
    ]


class DelugeDownloadClient(TorrentDownloadClient):
    """
    Deluge download client.
    """

    type: Literal["deluge"] = "deluge"
    """
    Type value associated with this kind of download client.
    """

    host: NonEmptyStr
    """
    Deluge host name.
    """

    port: Port = 8112  # type: ignore[assignment]
    """
    Download client access port.
    """

    use_ssl: bool = False
    """
    Use a secure connection when connecting to the download client.
    """

    url_base: Optional[str] = None
    """
    Adds a prefix to the Deluge JSON URL, e.g. `http://[host]:[port]/[url_base]/json`.
    """

    password: Password
    """
    Password to use to authenticate the download client user.
    """

    category: Optional[str] = "tv-sonarr"
    """
    Associate media from Sonarr with a category.

    Adding a category specific to Sonarr avoids conflicts with unrelated non-Sonarr downloads.
    Using a category is optional, but strongly recommended.
    """

    postimport_category: Optional[str] = None
    """
    Category for Sonarr to set after it has imported the download.

    Sonarr will not remove torrents in that category even if seeding has finished.
    Leave blank, set to `null` or undefined to keep the same category.
    """

    recent_priority: DelugePriority = DelugePriority.last
    """
    Priority to use when grabbing episodes that aired within the last 14 days.

    Values:

    * `last`
    * `first`
    """

    older_priority: DelugePriority = DelugePriority.last
    """
    Priority to use when grabbing episodes that aired over 14 days ago.

    Values:

    * `last`
    * `first`
    """

    _implementation_name: str = "Deluge"
    _implementation: str = "Deluge"
    _config_contract: str = "DelugeSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        (
            "url_base",
            "urlBase",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("password", "password", {"is_field": True}),
        (
            "category",
            "tvCategory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        (
            "postimport_category",
            "tvImportedCategory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("recent_priority", "recentTvPriority", {"is_field": True}),
        ("older_priority", "olderTvPriority", {"is_field": True}),
    ]


class DownloadstationTorrentDownloadClient(TorrentDownloadClient):
    """
    Download client which uses torrents via Download Station.
    """

    type: Literal["downloadstation-torrent"] = "downloadstation-torrent"
    """
    Type value associated with this kind of download client.
    """

    host: NonEmptyStr
    """
    Download Station host name.
    """

    port: Port = 5000  # type: ignore[assignment]
    """
    Download client access port.
    """

    use_ssl: bool = False
    """
    Use a secure connection when connecting to the download client.
    """

    username: NonEmptyStr
    """
    User name to use when authenticating with the download client.
    """

    password: Password
    """
    Password to use to authenticate the download client user.
    """

    category: Optional[str] = None
    """
    Associate media from Sonarr with a category.
    Creates a `[category]` subdirectory in the output directory.

    Adding a category specific to Sonarr avoids conflicts with unrelated non-Sonarr downloads.
    Using a category is optional, but strongly recommended.
    """

    directory: Optional[str] = None
    """
    Optional shared folder to put downloads into.

    Leave blank, set to `null` or undefined to use the default download client location.
    """

    _implementation_name: str = "Download Station"
    _implementation: str = "TorrentDownloadStation"
    _config_contract: str = "DownloadStationSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        ("username", "username", {"is_field": True}),
        ("password", "password", {"is_field": True}),
        (
            "category",
            "tvCategory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        (
            "category",
            "tvDirectory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
    ]


class FloodDownloadClient(TorrentDownloadClient):
    """
    Flood download client.
    """

    type: Literal["flood"] = "flood"
    """
    Type value associated with this kind of download client.
    """

    host: NonEmptyStr
    """
    Flood host name.
    """

    port: Port = 3000  # type: ignore[assignment]
    """
    Download client access port.
    """

    use_ssl: bool = False
    """
    Use a secure connection when connecting to the download client.
    """

    url_base: Optional[str] = None
    """
    Optionally adds a prefix to Flood API, such as `[protocol]://[host]:[port]/[url_base]api`.
    """

    username: NonEmptyStr
    """
    User name to use when authenticating with the download client.
    """

    password: Password
    """
    Password to use to authenticate the download client user.
    """

    destination: Optional[str] = None
    """
    Manually specified download destination.
    """

    flood_tags: Set[NonEmptyStr] = {"sonarr"}  # type: ignore[arg-type]
    """
    Initial tags of a download within Flood.

    To be recognized, a download must have all initial tags.
    This avoids conflicts with unrelated downloads.
    """

    postimport_tags: Set[NonEmptyStr] = set()
    """
    Tags to append within Flood after a download has been imported into Sonarr.
    """

    additional_tags: Set[FloodMediaTag] = set()
    """
    Adds properties of media as tags within Flood.
    """

    start_on_add: bool = True
    """
    Immediately start download once the media has been added to the client.
    """

    _implementation_name: str = "Flood"
    _implementation: str = "Flood"
    _config_contract: str = "FloodSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        (
            "url_base",
            "urlBase",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("username", "username", {"is_field": True}),
        ("password", "password", {"is_field": True}),
        (
            "destination",
            "destination",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("flood_tags", "tags", {"is_field": True, "encoder": sorted}),
        ("postimport_tags", "postImportTags", {"is_field": True, "encoder": sorted}),
        ("recent_priority", "recentTvPriority", {"is_field": True}),
        (
            "additional_tags",
            "additionalTags",
            {
                "is_field": True,
                "decoder": lambda v: set(FloodMediaTag(t) for t in v),
                "encoder": lambda v: sorted(t.value for t in v),
            },
        ),
        ("start_on_add", "startOnAdd", {"is_field": True}),
    ]


class HadoukenDownloadClient(TorrentDownloadClient):
    """
    Hadouken download client.
    """

    type: Literal["hadouken"] = "hadouken"
    """
    Type value associated with this kind of download client.
    """

    host: NonEmptyStr
    """
    Hadouken host name.
    """

    port: Port = 7070  # type: ignore[assignment]
    """
    Download client access port.
    """

    use_ssl: bool = False
    """
    Use a secure connection when connecting to the download client.
    """

    url_base: Optional[str] = None
    """
    Adds a prefix to the Hadouken url, e.g. `http://[host]:[port]/[url_base]/api`.
    """

    username: NonEmptyStr
    """
    User name to use when authenticating with the download client.
    """

    password: Password
    """
    Password to use to authenticate the download client user.
    """

    category: NonEmptyStr = "sonarr-tv"  # type: ignore[assignment]
    """
    Associate media from Sonarr with a category.

    Adding a category specific to Sonarr avoids conflicts with unrelated non-Sonarr downloads.
    Using a category is optional, but strongly recommended.
    """

    _implementation_name: str = "Hadouken"
    _implementation: str = "Hadouken"
    _config_contract: str = "HadoukenSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        (
            "url_base",
            "urlBase",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("username", "username", {"is_field": True}),
        ("password", "password", {"is_field": True}),
        ("category", "category", {"is_field": True}),
    ]


class QbittorrentDownloadClient(TorrentDownloadClient):
    """
    qBittorrent download client.
    """

    type: Literal["qbittorrent"] = "qbittorrent"
    """
    Type value associated with this kind of download client.
    """

    host: NonEmptyStr
    """
    qBittorrent host name.
    """

    port: Port = 8080  # type: ignore[assignment]
    """
    Download client access port.
    """

    use_ssl: bool = False
    """
    Use a secure connection when connecting to the download client.
    """

    url_base: Optional[str] = None
    """
    Adds a prefix to the qBittorrent URL, e.g. `http://[host]:[port]/[url_base]/api`.
    """

    username: NonEmptyStr
    """
    User name to use when authenticating with the download client.
    """

    password: Password
    """
    Password to use to authenticate the download client user.
    """

    category: Optional[str] = "tv-sonarr"
    """
    Associate media from Sonarr with a category.

    Adding a category specific to Sonarr avoids conflicts with unrelated non-Sonarr downloads.
    Using a category is optional, but strongly recommended.
    """

    postimport_category: Optional[str] = None
    """
    Category for Sonarr to set after it has imported the download.

    Sonarr will not remove torrents in that category even if seeding has finished.
    Leave blank, set to `null` or undefined to keep the same category.
    """

    recent_priority: QbittorrentPriority = QbittorrentPriority.last
    """
    Priority to use when grabbing episodes that aired within the last 14 days.

    Values:

    * `last`
    * `first`
    """

    older_priority: QbittorrentPriority = QbittorrentPriority.last
    """
    Priority to use when grabbing episodes that aired over 14 days ago.

    Values:

    * `last`
    * `first`
    """

    initial_state: QbittorrentState = QbittorrentState.start
    """
    Initial state for torrents added to qBittorrent.

    Note that forced torrents do not abide by seed restrictions.
    """

    sequential_order: bool = False
    """
    Download files in sequential order.

    This option requires qBittorrent version 4.1.0 or later.
    """

    first_and_last_first: bool = False
    """
    Download first and last pieces of a file first.

    This option requires qBittorrent version 4.1.0 or later.
    """

    _implementation_name: str = "qBittorrent"
    _implementation: str = "QBittorrent"
    _config_contract: str = "QBittorrentSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        (
            "url_base",
            "urlBase",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("username", "username", {"is_field": True}),
        ("password", "password", {"is_field": True}),
        (
            "category",
            "tvCategory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        (
            "postimport_category",
            "tvImportedCategory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("recent_priority", "recentTvPriority", {"is_field": True}),
        ("older_priority", "olderTvPriority", {"is_field": True}),
        ("initial_state", "initialState", {"is_field": True}),
        ("sequential_order", "sequentialOrder", {"is_field": True}),
        ("first_and_last_first", "firstAndLast", {"is_field": True}),
    ]


class RtorrentDownloadClient(TorrentDownloadClient):
    """
    RTorrent (ruTorrent) download client.
    """

    type: Literal["rtorrent", "rutorrent"] = "rtorrent"
    """
    Type value associated with this kind of download client.
    """

    host: NonEmptyStr
    """
    RTorrent host name.
    """

    port: Port = 8080  # type: ignore[assignment]
    """
    Download client access port.
    """

    use_ssl: bool = False
    """
    Use a secure connection when connecting to the download client.
    """

    url_base: NonEmptyStr = "RPC2"  # type: ignore[assignment]
    """
    Path to the XMLRPC endpoint, e.g. `http(s)://[host]:[port]/[url_base]`.

    When using RTorrent this usually is `RPC2` or `plugins/rpc/rpc.php`.
    """

    username: NonEmptyStr
    """
    User name to use when authenticating with the download client.
    """

    password: Password
    """
    Password to use to authenticate the download client user.
    """

    category: Optional[str] = "tv-sonarr"
    """
    Associate media from Sonarr with a category.

    Adding a category specific to Sonarr avoids conflicts with unrelated non-Sonarr downloads.
    Using a category is optional, but strongly recommended.
    """

    postimport_category: Optional[str] = None
    """
    Category for Sonarr to set after it has imported the download.

    Sonarr will not remove torrents in that category even if seeding has finished.
    Leave blank, set to `null` or undefined to keep the same category.
    """

    directory: Optional[str] = None
    """
    Optional shared folder to put downloads into.

    Leave blank, set to `null` or undefined to use the default download client location.
    """

    recent_priority: RtorrentPriority = RtorrentPriority.normal
    """
    Priority to use when grabbing episodes that aired within the last 14 days.

    Values:

    * `verylow`
    * `low`
    * `normal`
    * `high`
    """

    older_priority: RtorrentPriority = RtorrentPriority.normal
    """
    Priority to use when grabbing episodes that aired over 14 days ago.

    Values:

    * `verylow`
    * `low`
    * `normal`
    * `high`
    """

    add_stopped: bool = False
    """
    Enabling will add torrents and magnets to RTorrent in a stopped state.

    This may break magnet files.
    """

    _implementation_name: str = "rTorrent"
    _implementation: str = "RTorrent"
    _config_contract: str = "RTorrentSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        ("url_base", "urlBase", {"is_field": True}),
        ("username", "username", {"is_field": True}),
        ("password", "password", {"is_field": True}),
        (
            "category",
            "tvCategory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        (
            "postimport_category",
            "tvImportedCategory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        (
            "directory",
            "tvDirectory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("recent_priority", "recentTvPriority", {"is_field": True}),
        ("older_priority", "olderTvPriority", {"is_field": True}),
        ("add_stopped", "addStopped", {"is_field": True}),
    ]


class TorrentBlackholeDownloadClient(TorrentDownloadClient):
    """
    Torrent Blackhole download client.
    """

    type: Literal["torrent-blackhole"] = "torrent-blackhole"
    """
    Type value associated with this kind of download client.
    """

    torrent_folder: NonEmptyStr
    """
    Folder in which Sonarr will store `.torrent` files.
    """

    watch_folder: NonEmptyStr
    """
    Folder from which Sonarr should import completed downloads.
    """

    save_magnet_files: bool = False
    """
    Save the magnet link if no `.torrent` file is available.

    Only useful if the download client supports magnets saved to a file.
    """

    magnet_file_extension: NonEmptyStr = ".magnet"  # type: ignore[assignment]
    """
    Extension to use for magnet links.
    """

    read_only: bool = True
    """
    Instead of moving files, this will instruct Sonarr to copy or hard link
    (depending on settings/system configuration).
    """

    _implementation_name: str = "Torrent Blackhole"
    _implementation: str = "TorrentBlackhole"
    _config_contract: str = "TorrentBlackholeSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        ("url_base", "urlBase", {"is_field": True}),
        ("username", "username", {"is_field": True}),
        ("password", "password", {"is_field": True}),
        (
            "category",
            "tvCategory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        (
            "postimport_category",
            "tvImportedCategory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        (
            "directory",
            "tvDirectory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("recent_priority", "recentTvPriority", {"is_field": True}),
        ("older_priority", "olderTvPriority", {"is_field": True}),
        ("add_stopped", "addStopped", {"is_field": True}),
    ]


class TransmissionDownloadClientBase(TorrentDownloadClient):
    """
    Configuration options common to both Transmission and Vuze download clients:
    """

    host: NonEmptyStr
    """
    Transmission/Vuze host name.
    """

    port: Port = 9091  # type: ignore[assignment]
    """
    Download client access port.
    """

    use_ssl: bool = False
    """
    Use a secure connection when connecting to the download client.
    """

    url_base: NonEmptyStr = "/transmission/"  # type: ignore[assignment]
    """
    Adds a prefix to the Transmission/Vuze RPC url, e.g.`http://[host]:[port][url_base]rpc`.

    This is set by default in most clients to `/transmission/`.
    """

    username: Optional[str] = None
    """
    User name to use when authenticating with the download client, if required.
    """

    password: Optional[Password] = None
    """
    Password to use to authenticate the download client user, if required.
    """

    category: Optional[str] = None
    """
    Associate media from Sonarr with a category.
    Creates a `[category]` subdirectory in the output directory.

    Adding a category specific to Sonarr avoids conflicts with unrelated non-Sonarr downloads.
    Using a category is optional, but strongly recommended.
    """

    directory: Optional[str] = None
    """
    Optional shared folder to put downloads into.

    Leave blank, set to `null` or undefined to use the default download client location.
    """

    recent_priority: TransmissionPriority = TransmissionPriority.last
    """
    Priority to use when grabbing episodes that aired within the last 14 days.

    Values:

    * `last`
    * `first`
    """

    older_priority: TransmissionPriority = TransmissionPriority.last
    """
    Priority to use when grabbing episodes that aired over 14 days ago.

    Values:

    * `last`
    * `first`
    """

    add_paused: bool = False
    """
    Add media to the download client in the Paused state.
    """

    _config_contract: str = "TransmissionSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        ("url_base", "urlBase", {"is_field": True}),
        (
            "username",
            "username",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("password", "password", {"is_field": True, "field_default": None}),
        (
            "category",
            "tvCategory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        (
            "directory",
            "tvDirectory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("recent_priority", "recentTvPriority", {"is_field": True}),
        ("older_priority", "olderTvPriority", {"is_field": True}),
        ("add_paused", "addPaused", {"is_field": True}),
    ]


class TransmissionDownloadClient(TransmissionDownloadClientBase):
    """
    Tramsmission download client.
    """

    type: Literal["transmission"] = "transmission"
    """
    Type value associated with this kind of download client.
    """

    _implementation_name: str = "Transmission"
    _implementation: str = "Transmission"


class VuzeDownloadClient(TransmissionDownloadClientBase):
    """
    Vuze download client.
    """

    type: Literal["vuze"] = "vuze"
    """
    Type value associated with this kind of download client.
    """

    _implementation_name: str = "Vuze"
    _implementation: str = "Vuze"


class UtorrentDownloadClient(TorrentDownloadClient):
    """
    uTorrent download client.
    """

    type: Literal["utorrent"] = "utorrent"
    """
    Type value associated with this kind of download client.
    """

    host: NonEmptyStr
    """
    uTorrent host name.
    """

    port: Port = 8080  # type: ignore[assignment]
    """
    Download client access port.
    """

    use_ssl: bool = False
    """
    Use a secure connection when connecting to the download client.
    """

    url_base: Optional[str] = None
    """
    Adds a prefix to the uTorrent URL, e.g. `http://[host]:[port]/[url_base]/api`.
    """

    username: NonEmptyStr
    """
    User name to use when authenticating with the download client.
    """

    password: Password
    """
    Password to use to authenticate the download client user.
    """

    category: Optional[str] = "tv-sonarr"
    """
    Associate media from Sonarr with a category.

    Adding a category specific to Sonarr avoids conflicts with unrelated non-Sonarr downloads.
    Using a category is optional, but strongly recommended.
    """

    postimport_category: Optional[str] = None
    """
    Category for Sonarr to set after it has imported the download.

    Sonarr will not remove torrents in that category even if seeding has finished.
    Leave blank, set to `null` or undefined to keep the same category.
    """

    recent_priority: UtorrentPriority = UtorrentPriority.last
    """
    Priority to use when grabbing episodes that aired within the last 14 days.

    Values:

    * `last`
    * `first`
    """

    older_priority: UtorrentPriority = UtorrentPriority.last
    """
    Priority to use when grabbing episodes that aired over 14 days ago.

    Values:

    * `last`
    * `first`
    """

    initial_state: UtorrentState = UtorrentState.start
    """
    Initial state for torrents added to uTorrent.
    """

    _implementation_name: str = "uTorrent"
    _implementation: str = "UTorrent"
    _config_contract: str = "UTorrentSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {"is_field": True}),
        ("port", "port", {"is_field": True}),
        ("use_ssl", "useSsl", {"is_field": True}),
        (
            "url_base",
            "urlBase",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("username", "username", {"is_field": True}),
        ("password", "password", {"is_field": True}),
        (
            "category",
            "tvCategory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        (
            "postimport_category",
            "tvImportedCategory",
            {"is_field": True, "decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("recent_priority", "recentTvPriority", {"is_field": True}),
        ("older_priority", "olderTvPriority", {"is_field": True}),
        ("initial_state", "initialState", {"is_field": True}),
    ]


DOWNLOADLCLIENT_TYPES: Tuple[Type[DownloadClient], ...] = (
    DownloadstationUsenetDownloadClient,
    NzbgetDownloadClient,
    NzbvortexDownloadClient,
    PneumaticDownloadClient,
    SabnzbdDownloadClient,
    UsenetBlackholeDownloadClient,
    Aria2DownloadClient,
    DelugeDownloadClient,
    DownloadstationTorrentDownloadClient,
    FloodDownloadClient,
    HadoukenDownloadClient,
    QbittorrentDownloadClient,
    RtorrentDownloadClient,
    TorrentBlackholeDownloadClient,
    TransmissionDownloadClient,
    UtorrentDownloadClient,
    VuzeDownloadClient,
)

DOWNLOADCLIENT_TYPE_MAP: Dict[str, Type[DownloadClient]] = {
    downloadclient_type._implementation: downloadclient_type
    for downloadclient_type in DOWNLOADLCLIENT_TYPES
}
